# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine-client setup code."""

import errno
import json
import tempfile
import textwrap
from pathlib import Path
from unittest import mock

import responses
from configobj import ConfigObj

from debusine.client.config import KNOWN_SERVERS, ServerInfo
from debusine.client.models import (
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
)
from debusine.client.setup import (
    DebusineClientConfig,
    ServerConfigEditor,
    ServerSelector,
    TokenFetchFailed,
    fetch_token,
    setup_server,
)
from debusine.client.tests.utils import BufferConsole, MockTable
from debusine.test import TestCase
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT

FEEDBACK_CONFIG_SAVED = "👍 Configuration saved"
FEEDBACK_TOKEN_ACQUIRED = "✅ Confirmation confirmed, token acquired"


class ServerInfoTests(TestCase):
    """Tests for :py:class:`ServerInfo`."""

    def test_from_config(self) -> None:
        """Test from_config."""
        workdir = self.create_temporary_directory()
        path = workdir / "debusine.ini"
        desc = f"Configured in {path}"
        for name, cfg, info in (
            (
                "example",
                {"api-url": "http://example.org/api/", "scope": "debusine"},
                ServerInfo(
                    name="example",
                    api_url="http://example.org/api/",
                    desc=desc,
                    scope="debusine",
                    default_workspace="System",
                ),
            ),
            (
                "example",
                {
                    "api-url": "http://example.org/api/",
                    "scope": "debusine",
                    "default-workspace": "dw",
                },
                ServerInfo(
                    name="example",
                    api_url="http://example.org/api/",
                    desc=desc,
                    scope="debusine",
                    default_workspace="dw",
                ),
            ),
            (
                "debian",
                {"scope": "debusine"},
                ServerInfo(
                    name="debian",
                    api_url="https://debusine.debian.net/api/",
                    desc=desc,
                    scope="debusine",
                    default_workspace="developers",
                ),
            ),
            (
                "empty",
                {},
                ServerInfo(
                    name="empty",
                    api_url="https://empty/api/",
                    desc=desc,
                    scope="debusine",
                    default_workspace="System",
                ),
            ),
        ):
            initial_cfg = ConfigObj({f"server:{name}": cfg})
            initial_cfg.filename = path.as_posix()
            initial_cfg.write()
            loaded_cfg = ConfigObj(path.as_posix())
            loaded_info = ServerInfo.from_config(
                name, loaded_cfg[f"server:{name}"], path
            )
            assert loaded_info == info

    def test_from_config_invalid_timeout_logs_warning(self) -> None:
        """Test that invalid http-timeout logs a warning and uses default."""
        workdir = self.create_temporary_directory()
        path = workdir / "debusine.ini"

        cfg = {
            "api-url": "http://example.org/api/",
            "scope": "debusine",
            "http-timeout": "invalid_value",
        }

        initial_cfg = ConfigObj({"server:test": cfg})
        initial_cfg.filename = path.as_posix()
        initial_cfg.write()

        with self.assertLogs("debusine.client.config", level="WARNING") as cm:
            loaded_cfg = ConfigObj(path.as_posix())
            loaded_info = ServerInfo.from_config(
                "test", loaded_cfg["server:test"], path
            )

        self.assertEqual(loaded_info.http_timeout, DEFAULT_REQUEST_TIMEOUT)
        self.assertTrue(
            any(
                "Invalid http-timeout value in configuration" in output
                for output in cm.output
            )
        )  # pragma: no cover (can be removed post-trixie)

    def test_from_config_with_timeout(self) -> None:
        """Test from_config with http-timeout and defaults."""
        workdir = self.create_temporary_directory()
        path = workdir / "debusine.ini"

        server_name = "test-server"

        # (Input Value from .ini, Expected Python object)
        test_cases = (
            ("30", 30),
            ("10, 60", (10, 60)),
            ("invalid", DEFAULT_REQUEST_TIMEOUT),
            (None, DEFAULT_REQUEST_TIMEOUT),
        )

        for raw_val, expected in test_cases:
            with self.subTest(raw_val=raw_val):
                cfg = {
                    "api-url": "http://example.org/api/",
                    "scope": "debusine",
                }

                # Only add the key if raw_val is not None
                if raw_val is not None:
                    cfg["http-timeout"] = raw_val

                # Write to temporary file
                initial_cfg = ConfigObj({f"server:{server_name}": cfg})
                initial_cfg.filename = path.as_posix()
                initial_cfg.write()

                # Load back and verify
                loaded_cfg = ConfigObj(path.as_posix())
                loaded_info = ServerInfo.from_config(
                    server_name, loaded_cfg[f"server:{server_name}"], path
                )

                self.assertEqual(
                    loaded_info.http_timeout,
                    expected,
                    f"Failed for input: {raw_val}",
                )

    def test_from_string_known(self) -> None:
        """Test from_string with a known name or hostname."""
        for arg, name in (
            # Match by name
            ("debian", "debian"),
            ("localhost", "localhost"),
            # Match by hostname
            ("debusine.debian.net", "debian"),
            ("localhost:8000", "localhost"),
            (":8000", "localhost"),
            # Match by URL hostname
            ("https://debusine.debian.net/", "debian"),
            ("http://localhost:8000/api/", "localhost"),
        ):
            with self.subTest(arg=arg):
                self.assertIs(
                    ServerInfo.from_string(arg, desc="test"),
                    KNOWN_SERVERS[name],
                )

    def test_from_string(self) -> None:
        """Test building a ServerInfo from a string argument."""
        for server, server_name, api_url in (
            # Bare name
            ("example", "example", "https://example/api/"),
            # Domain name
            (
                "example.org",
                "example.org",
                "https://example.org/api/",
            ),
            # Domain name and port
            (
                "example.org:8000",
                "example.org",
                "https://example.org:8000/api/",
            ),
            # Base URL
            (
                "https://example.org",
                "example.org",
                "https://example.org/api/",
            ),
            # API URL without trailing slash
            (
                "https://example.org/api",
                "example.org",
                "https://example.org/api/",
            ),
            # API URL with trailing slash
            (
                "https://example.org/api/",
                "example.org",
                "https://example.org/api/",
            ),
            # Subdir URL without trailing slash
            (
                "https://example.org/debusine",
                "example.org",
                "https://example.org/debusine/api/",
            ),
            # Subdir URL with trailing slash
            (
                "https://example.org/debusine/",
                "example.org",
                "https://example.org/debusine/api/",
            ),
        ):
            with self.subTest(server=server):
                self.assertEqual(
                    ServerInfo.from_string(server, "desc"),
                    ServerInfo(
                        name=server_name,
                        api_url=api_url,
                        scope="debusine",
                        desc="desc",
                        default_workspace="System",
                    ),
                )

    def test_lint_api_url(self) -> None:
        """Test linting of API URLs."""
        for api_url, messages in (
            (
                "https://localhost/api/",
                [
                    "Server url is an [bold]https[/] url,"
                    " should it be [bold]http[/]?"
                ],
            ),
            (
                "http://example.org/api/",
                [
                    "Server url is an [bold]http[/] url,"
                    " should it be [bold]https[/]?"
                ],
            ),
            ("https://example.org/api", []),
            (
                "https://example.org/other/",
                ["API URL does not end in /api or /api/"],
            ),
            (
                "http://example.org/",
                [
                    "Server url is an [bold]http[/] url,"
                    " should it be [bold]https[/]?",
                    "API URL does not end in /api or /api/",
                ],
            ),
        ):
            with self.subTest(api_url=api_url):
                info = ServerInfo(
                    name="test", desc="test", api_url=api_url, scope="test"
                )
                self.assertEqual(info.lint(), messages)

    def _fetch_token(
        self,
        nonce: str = "12345678",
        challenge: str = "correct horse battery staple",
    ) -> tuple[str | None, list[str], TokenFetchFailed | None]:
        exception: TokenFetchFailed | None
        server = KNOWN_SERVERS["debian"]
        console = BufferConsole()
        with (
            mock.patch(
                "debusine.client.setup.secrets.token_urlsafe",
                return_value=nonce,
            ),
            mock.patch(
                "debusine.client.setup.xkcd_password.generate_xkcdpassword",
                return_value=challenge,
            ),
            mock.patch(
                "debusine.client.setup.platform.node",
                return_value="hostname",
            ),
        ):
            try:
                result = fetch_token(console, server)
            except TokenFetchFailed as e:
                exception = e
                result = None
            else:
                exception = None

        # A proper payload was sent
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        assert call.request.body is not None
        self.assertEqual(
            call.request.headers["Content-Type"], "application/json"
        )
        self.assertEqual(
            json.loads(call.request.body),
            EnrollPayload(
                nonce=nonce,
                challenge=challenge,
                hostname="hostname",
                scope=server.scope,
            ).model_dump(),
        )

        output_lines = console.output.getvalue().splitlines()

        confirm_url = f"https://debusine.debian.net/-/enroll/confirm/{nonce}"
        self.assertEqual(
            output_lines[0],
            f"👉 Please visit {confirm_url} to confirm registration",
        )

        self.assertEqual(
            output_lines[1],
            f'👉 Make sure the page mentions the passphrase: "{challenge}"',
        )

        return result, output_lines[2:], exception

    @responses.activate
    def test_fetch_token_confirm(self) -> None:
        """Test getting a token from the server."""
        nonce = "12345678"
        challenge = "correct horse battery staple"
        token = "a" * 32
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            json=EnrollConfirmPayload(
                outcome=EnrollOutcome.CONFIRM, token=token
            ).model_dump(),
        )

        # Fetch the token
        result, feedback, exception = self._fetch_token(nonce, challenge)
        self.assertEqual(result, token)
        self.assertIsNone(exception)

        # URL and challenge were printed in console
        self.assertEqual(
            feedback, ["✅ Confirmation confirmed, token acquired"]
        )

    @responses.activate
    def test_fetch_token_cancel(self) -> None:
        """Test getting a token from the server."""
        nonce = "12345678"
        challenge = "correct horse battery staple"
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            json=EnrollConfirmPayload(
                outcome=EnrollOutcome.CANCEL
            ).model_dump(),
        )

        # Fetch the token
        result, feedback, exception = self._fetch_token(nonce, challenge)
        self.assertIsNone(result)
        self.assertIsNone(exception)

        # URL and challenge were printed in console
        self.assertEqual(feedback, ["❗ Confirmation cancelled"])

    @responses.activate
    def test_fetch_token_server_not_found(self) -> None:
        """Test fetch_token encountering a network error."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            body=OSError(errno.ESRCH),
        )

        # Fetch the token
        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(str(exception), "request to server failed: 3")
        self.assertEqual(feedback, [])

    @responses.activate
    def test_fetch_token_server_problemresponse(self) -> None:
        """Test fetch_token getting a ProblemResponse."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            status=404,
            json={"title": "fake server error"},
        )

        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(
            str(exception),
            "The server returned status code 404: fake server error",
        )
        self.assertEqual(feedback, [])

    @responses.activate
    def test_fetch_token_non_problemresponse_json(self) -> None:
        """Test a JSON server error but not a ProblemResponse."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            status=500,
            json={"mischief": True},
        )

        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(
            str(exception),
            """The server returned status code 500: {"mischief": true}""",
        )
        self.assertEqual(feedback, [])

    @responses.activate
    def test_fetch_token_server_error(self) -> None:
        """Test fetch_token getting a server error."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            status=500,
            body="fake internal server error",
        )

        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(
            str(exception),
            "The server returned status code 500: fake internal server error",
        )
        self.assertEqual(feedback, [])

    @responses.activate
    def test_fetch_token_server_badjson(self) -> None:
        """Test fetch_token getting invalid json."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            body="{",
            content_type="application/json",
        )

        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(str(exception), "Invalid JSON in response")
        self.assertEqual(feedback, [])

    @responses.activate
    def test_fetch_token_server_badpayload(self) -> None:
        """Test fetch_token getting a bad payload."""
        responses.add(
            responses.POST,
            "https://debusine.debian.net/api/enroll/",
            json={"mischief": True},
        )

        result, feedback, exception = self._fetch_token()
        self.assertIsNone(result)
        self.assertEqual(str(exception), "Invalid response payload")
        self.assertEqual(feedback, [])


class DebusineClientConfigTests(TestCase):
    """Tests for :py:class:`DebusineClientConfig`."""

    def make_config(
        self,
        path_or_contents: Path | str = Path("/dev/null"),
        selected: str | None = "debian",
    ) -> DebusineClientConfig:
        """Create a DebusineClientConfig object for testing."""
        if isinstance(path_or_contents, str):
            path = self.create_temporary_file()
            path.write_text(textwrap.dedent(path_or_contents))
        else:
            path = path_or_contents
        config = DebusineClientConfig(path)
        if selected is not None:
            config.select(selected, "test")
        return config

    def test_default_server_unset(self) -> None:
        """Test default_server when not configured."""
        config = DebusineClientConfig(Path("/dev/null"))
        self.assertIsNone(config.default_server)

    def test_default_server_set_existing(self) -> None:
        """Test default_server when set to an existing server."""
        config = self.make_config(
            """
            [General]
            default-server = debian
            [server:debian]
            """,
        )
        self.assertEqual(config.default_server, "debian")

    def test_default_server_set_not_existing(self) -> None:
        """Test default_server when set to a non-existing server."""
        config = self.make_config(
            """
            [General]
            default-server = test
            """,
            selected=None,
        )
        self.assertEqual(config.default_server, "test")

    def test_set_default(self) -> None:
        config = self.make_config(
            """
            [General]
            default-server = debian
            [server:debian]
            """,
            selected="localhost",
        )
        self.assertEqual(config.default_server, "debian")
        self.assertEqual(config.server.name, "localhost")
        config.set_default()
        self.assertEqual(config.default_server, "localhost")
        config.save()
        self.assertEqual(
            config.path.read_text(),
            textwrap.dedent(
                """
                [General]
                default-server = localhost
                [server:debian]
                [server:localhost]
                api-url = http://localhost:8000/api/
                scope = debusine
                default-workspace = Playground
                """
            ),
        )

    def test_save_default_never_set(self) -> None:
        """Test saving when default server has never been set."""
        config = self.make_config("[server:debian]")
        config.save()
        self.assertEqual(config.path.read_text(), "[server:debian]\n")

    def test_save_default_initially_unset(self) -> None:
        """Test saving when default server is initially unset."""
        config = self.make_config("[server:debian]")
        config.set_default()
        config.save()
        self.assertEqual(
            config.path.read_text(),
            textwrap.dedent(
                """
        [server:debian]
        [General]
        default-server = debian
        """
            ).lstrip(),
        )

    def test_save_default_changed(self) -> None:
        """Test saving when default server is initially unset."""
        config = self.make_config(
            """
            [General]
            default-server=debian
            [server:debian]
            """,
            selected="localhost",
        )
        config.set_default()
        config.save()
        self.assertEqual(
            config.path.read_text(),
            textwrap.dedent(
                """
        [General]
        default-server = localhost
        [server:debian]
        [server:localhost]
        api-url = http://localhost:8000/api/
        scope = debusine
        default-workspace = Playground
        """
            ),
        )

    def test_save_default_unchanged(self) -> None:
        """Test saving when default server is unchanged."""
        config = self.make_config(
            """
            [General]
            default-server=debian
            [server:debian]
            """,
            selected="localhost",
        )
        config.save()
        self.assertEqual(
            config.path.read_text(),
            textwrap.dedent(
                """
                [General]
                default-server = debian
                [server:debian]
                [server:localhost]
                api-url = http://localhost:8000/api/
                scope = debusine
                default-workspace = Playground
                """,
            ),
        )

    def test_save(self) -> None:
        """Test saving the configuration."""

        def body(text: str) -> str:
            return textwrap.dedent(text).lstrip()

        for initial, server, changed, expected in (
            # Fill with defaults
            (
                "",
                "localhost",
                {},
                body(
                    """
                    [server:localhost]
                    api-url = http://localhost:8000/api/
                    scope = debusine
                    default-workspace = Playground
                    """
                ),
            ),
            # Add a section from defaults
            (
                body(
                    """
                    # Debian account
                    [server:debian]
                    api-url=http://debusine.debian.net/api/
                    scope=debian
                    # Test comment
                    token=12345678
                    """
                ),
                "localhost",
                {},
                body(
                    """
                    # Debian account
                    [server:debian]
                    api-url = http://debusine.debian.net/api/
                    scope = debian
                    # Test comment
                    token = 12345678
                    [server:localhost]
                    api-url = http://localhost:8000/api/
                    scope = debusine
                    default-workspace = Playground
                    """
                ),
            ),
            # Fetch a token for an existing section
            (
                body(
                    """
                    # Debian account
                    [server:debian]
                    api-url=http://debusine.debian.net/api/
                    scope=debian
                    """
                ),
                "debian",
                {"api_token": "12345678"},
                body(
                    """
                    # Debian account
                    [server:debian]
                    api-url = http://debusine.debian.net/api/
                    scope = debian
                    token = 12345678
                    """
                ),
            ),
            # Change API URL, preserving comments
            (
                body(
                    """
                    # Debian account
                    [server:localhost]
                    # Misses a port?
                    api-url=http://localhost/api/
                    scope=debian
                    """
                ),
                "localhost",
                {"api_url": "http://localhost:8000/api/"},
                body(
                    """
                    # Debian account
                    [server:localhost]
                    # Misses a port?
                    api-url = http://localhost:8000/api/
                    scope = debian
                    """
                ),
            ),
            # Clear API token
            (
                body(
                    """
                    [server:localhost]
                    api-url=http://localhost/api/
                    scope=debian
                    # Token value
                    token=12345678
                    """
                ),
                "localhost",
                {"api_token": None},
                body(
                    """
                    [server:localhost]
                    api-url = http://localhost/api/
                    scope = debian
                    """
                ),
            ),
        ):
            with (
                self.subTest(initial=initial, server=server, changed=changed),
                tempfile.NamedTemporaryFile("w+t") as cfg_file,
            ):
                cfg_file.write(initial)
                cfg_file.flush()
                cfg = DebusineClientConfig(Path(cfg_file.name))
                cfg.select(server, "desc")
                for k, v in changed.items():
                    getattr(cfg, f"set_{k}")(v)
                cfg.save()
                cfg_file.seek(0)
                self.assertEqual(cfg_file.read(), expected)

    def test_save_parent_dir_missing(self) -> None:
        """Test saving the configuration in a missing path."""
        workdir = self.create_temporary_directory()
        configdir = workdir / "client"
        configfile = configdir / "config.ini"
        self.assertFalse(configdir.exists())
        self.assertFalse(configfile.exists())
        cfg = self.make_config(configfile, "localhost")
        cfg.save()
        self.assertTrue(configdir.is_dir())
        self.assertTrue(configfile.is_file())

    def test_fetch_token(self) -> None:
        token = "aaaaaa"
        console = BufferConsole()
        cfg = self.make_config()
        with mock.patch(
            "debusine.client.setup.fetch_token", return_value=token
        ):
            self.assertTrue(cfg.fetch_token(console))
        self.assertEqual(cfg.server.api_token, token)
        self.assertEqual(console.output_lines(), [])

    def test_fetch_cancelled(self) -> None:
        console = BufferConsole()
        cfg = self.make_config()
        cfg.set_api_token("aaaaaa")
        with mock.patch("debusine.client.setup.fetch_token", return_value=None):
            self.assertTrue(cfg.fetch_token(console))
        self.assertIsNone(cfg.server.api_token)
        self.assertEqual(console.output_lines(), [])

    def test_fetch_token_error(self) -> None:
        token = "aaaaaa"
        console = BufferConsole()
        cfg = self.make_config()
        cfg.set_api_token(token)
        with mock.patch(
            "debusine.client.setup.fetch_token",
            side_effect=TokenFetchFailed("message"),
        ):
            self.assertFalse(cfg.fetch_token(console))
        self.assertEqual(console.output_lines(), ["❗ message"])
        self.assertEqual(cfg.server.api_token, token)

    def test_switch(self) -> None:
        """Test switching current server."""
        cfg = self.make_config(selected="debian")
        cfg.set_scope("testdebian")

        cfg.select("localhost", "test")
        self.assertEqual(cfg.server.name, "localhost")
        self.assertEqual(cfg.server.scope, "debusine")

        cfg.select("debian", "test")
        self.assertEqual(cfg.server.name, "debian")
        self.assertEqual(cfg.server.scope, "testdebian")

    def test_save_multiple(self) -> None:
        """Test saving multiple edited servers."""
        with tempfile.NamedTemporaryFile("w+t") as cfg_file:
            path = Path(cfg_file.name)
            path.write_text(
                textwrap.dedent(
                    """
            [server:test]
            api-url=https://test.example.org
            """
                ).lstrip()
            )

            cfg = self.make_config(path, selected="debian")
            cfg.set_scope("testdebian")
            cfg.select("test", "test")
            cfg.set_scope("testtest")
            cfg.save()

            saved = path.read_text()

        self.assertEqual(
            saved.splitlines(),
            [
                "[server:test]",
                "api-url = https://test.example.org",
                "scope = testtest",
                "[server:debian]",
                "api-url = https://debusine.debian.net/api/",
                "scope = testdebian",
                "default-workspace = developers",
            ],
        )

    def test_server_not_selected(self) -> None:
        """Accessing the current server when not selected raises an error."""
        config = DebusineClientConfig(Path("/dev/null"))
        with self.assertRaisesRegex(
            RuntimeError, "current server has not been selected"
        ):
            config.server


class ServerSelectorTests(TestCase):
    """Tests for :py:class:`ServerSelector`."""

    def test_cmdloop_no_entries(self) -> None:
        """Test cmdloop behaviour with no entries."""
        with (
            mock.patch(
                "debusine.client.dataentry.DataEntry.cmdloop"
            ) as cmdloop,
            mock.patch("debusine.client.setup.ServerSelector.do_new") as do_new,
        ):
            console = BufferConsole()
            selector = ServerSelector([], console=console)
            selector.cmdloop()
        cmdloop.assert_not_called()
        do_new.assert_called()

    def test_cmdloop_with_entries(self) -> None:
        """Test cmdloop behaviour with entries."""
        with (
            mock.patch(
                "debusine.client.dataentry.DataEntry.cmdloop"
            ) as cmdloop,
            mock.patch("debusine.client.setup.ServerSelector.do_new") as do_new,
        ):
            selector = ServerSelector(
                list(KNOWN_SERVERS.values()), console=BufferConsole()
            )
            selector.cmdloop()
        cmdloop.assert_called()
        do_new.assert_not_called()

    def test_do_new(self) -> None:
        """Test do_new."""
        for entered, stored in (
            ("", None),
            ("localhost", KNOWN_SERVERS["localhost"]),
            (
                "https://example.org",
                ServerInfo(
                    "example.org",
                    desc="Manually entered",
                    api_url="https://example.org/api/",
                    scope="debusine",
                ),
            ),
        ):
            with self.subTest(entered=entered):
                selector = ServerSelector(
                    [KNOWN_SERVERS["localhost"]], console=BufferConsole()
                )
                with mock.patch(
                    "debusine.client.dataentry.DataEntry.input_line",
                    return_value=entered,
                ):
                    selector.onecmd("new")
                self.assertEqual(selector.selected, stored)

    def test_table(self) -> None:
        """Test table display."""
        console = BufferConsole()
        selector = ServerSelector(
            list(KNOWN_SERVERS.values()),
            console=console,
            default="localhost",
        )
        with mock.patch("debusine.client.setup.Table", new=MockTable):
            selector.menu()

        table = console.tables[0]
        self.assertEqual(
            [c.args for c in table.rows],
            [
                ("1", "debian", "", "debusine.debian.net"),
                ("2", "freexian", "", "Freexian (E)LTS"),
                (
                    "3",
                    "localhost",
                    "yes",
                    "debusine-admin runserver for development",
                ),
            ],
        )

    def test_menu_prompt(self) -> None:
        """Test menu prompt."""
        for entries, expected in (
            ([], "new, quit:"),
            ([ServerInfo.from_string("a", "a")], "1, new, quit:"),
            (
                [ServerInfo.from_string(x, x) for x in ("a", "b", "c")],
                "1…3, new, quit:",
            ),
        ):
            with self.subTest(entries=repr(entries)):
                console = BufferConsole()
                selector = ServerSelector(entries, console=console)
                selector.menu()
                self.assertEqual(
                    console.output.getvalue().splitlines()[-1], expected
                )

    def test_enter_number(self) -> None:
        """Test entering a number to pick an entry."""
        selector = ServerSelector(
            list(KNOWN_SERVERS.values()), console=BufferConsole()
        )
        self.assertTrue(selector.onecmd("2"))
        self.assertIs(selector.selected, KNOWN_SERVERS["freexian"])

    def test_enter_number_out_of_range(self) -> None:
        """Test entering a number out of range."""
        console = BufferConsole()
        selector = ServerSelector(list(KNOWN_SERVERS.values()), console=console)
        for num in 0, len(KNOWN_SERVERS) + 1, len(KNOWN_SERVERS) + 10:
            with self.subTest(num=num):
                console.reset_output()
                self.assertFalse(selector.onecmd(str(num)))
                self.assertEqual(
                    console.output.getvalue(),
                    "Entry number must be between"
                    f" 1 and {len(KNOWN_SERVERS)}\n",
                )
                self.assertIsNone(selector.selected)

    def test_enter_invalid_command(self) -> None:
        """Test entering an invalid command."""
        console = BufferConsole()
        selector = ServerSelector(list(KNOWN_SERVERS.values()), console=console)
        for cmd in "-1", "14.3", "mischief":
            with self.subTest(cmd=cmd):
                console.reset_output()
                self.assertFalse(selector.onecmd(cmd))
                self.assertEqual(
                    console.output.getvalue(),
                    f"command not recognized: {cmd!r}\n",
                )
                self.assertIsNone(selector.selected)


class ServerConfigEditorTests(TestCase):
    """Tests for :py:class:`ServerConfigEditor`."""

    def _config(self) -> DebusineClientConfig:
        res = DebusineClientConfig(Path("/dev/null"))
        res.select("debian", "tests")
        return res

    def test_menu(self) -> None:
        """Test formatting the menu."""
        cfg = self._config()
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch("debusine.client.setup.Table", new=MockTable):
            editor.menu()
        table = console.tables[0]
        self.assertEqual(
            table.columns[0].args[0],
            "Configuration for debian"
            " [yellow not bold](default is [i]None[/])[/]",
        )
        self.assertEqual(
            [c.args for c in table.rows],
            [
                ("\\[server:[bold]debian[bold]]", ""),
                ("scope = [bold]debian[bold]", "([bold]sc[/]ope)"),
                (
                    "api-url = [bold]https://debusine.debian.net/api/[bold]",
                    "([bold]u[/]rl)",
                ),
                ("token = [italic]acquired on save[/]", "([bold]t[/]oken)"),
                (
                    "default-workspace = [bold]developers[bold]",
                    "([bold]w[/]orkspace)",
                ),
            ],
        )

        output = console.output.getvalue()
        self.assertEqual(
            output.splitlines()[-1],
            "Commands: url, token, scope, (make) default, save, quit:",
        )

    def test_menu_default_server(self) -> None:
        """Test menu showing the default server."""
        cfg = self._config()
        cfg.set_default()
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch("debusine.client.setup.Table", new=MockTable):
            editor.menu()
        table = console.tables[0]
        self.assertEqual(
            table.columns[0].args[0],
            "Configuration for debian [green not bold](default server)[/]",
        )

    def test_menu_with_token(self) -> None:
        """Test formatting the menu when the token is set."""
        cfg = self._config()
        cfg.set_api_token("12345678")
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        editor.menu()
        self.assertIn("token = present", console.output.getvalue())

    def test_menu_linter(self) -> None:
        """Ensure linter messages appear in the menu."""
        cfg = self._config()
        cfg.set_api_url("http://debusine.debian.net/api/")
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        editor.menu()
        self.assertIn(
            "👉 Server url is an http url, should it be https?",
            console.output_lines(),
        )

    def test_edit_scope(self) -> None:
        """Edit the scope."""
        cfg = self._config()
        self.assertEqual(cfg.server.scope, "debian")
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.dataentry.DataEntry.input_line",
            return_value="debusine",
        ) as input_line:
            editor.onecmd("scope")
        self.assertEqual(cfg.server.scope, "debusine")
        input_line.assert_called_with(
            "Enter scope", initial="debian", required=True
        )

    def test_edit_url(self) -> None:
        """Edit the api url."""
        old_url = "https://debusine.debian.net/api/"
        new_url = "https://debusine.debian.org/api/"
        cfg = self._config()
        self.assertEqual(cfg.server.api_url, old_url)
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.dataentry.DataEntry.input_line",
            return_value=new_url,
        ) as input_line:
            editor.onecmd("url")
        self.assertEqual(cfg.server.api_url, new_url)
        input_line.assert_called_with(
            "Enter API URL", initial=old_url, required=True
        )

    def test_edit_token(self) -> None:
        """Edit the token."""
        new_token = "12345678"
        cfg = self._config()
        self.assertIsNone(cfg.server.api_token)
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.dataentry.DataEntry.input_line",
            return_value=new_token,
        ) as input_line:
            editor.onecmd("token")
        self.assertEqual(cfg.server.api_token, new_token)
        input_line.assert_called_with(
            "Enter (or clear) API token", initial=None
        )

    def test_edit_default_workspace(self) -> None:
        """Edit the api url."""
        old_default_workspace = "developers"
        new_default_workspace = "Changed"
        cfg = self._config()
        self.assertEqual(cfg.server.default_workspace, old_default_workspace)
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.dataentry.DataEntry.input_line",
            return_value=new_default_workspace,
        ) as input_line:
            editor.onecmd("workspace")
        self.assertEqual(cfg.server.default_workspace, new_default_workspace)
        input_line.assert_called_with(
            "Enter default workspace name",
            initial=old_default_workspace,
            required=True,
        )

    def test_set_default(self) -> None:
        """Set the current server as default."""
        cfg = self._config()
        self.assertIsNone(cfg.default_server)
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        editor.onecmd("default")
        self.assertEqual(cfg.default_server, "debian")

    def test_clear_token(self) -> None:
        """Clear the token."""
        old_token = "12345678"
        cfg = self._config()
        cfg.set_api_token(old_token)
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.dataentry.DataEntry.input_line",
            return_value="",
        ) as input_line:
            editor.onecmd("token")
        self.assertIsNone(cfg.server.api_token)
        input_line.assert_called_with(
            "Enter (or clear) API token", initial=old_token
        )

    def test_quit(self) -> None:
        """Test quitting the editor."""
        cfg = self._config()
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        self.assertTrue(editor.onecmd("quit"))
        output = console.output.getvalue()
        self.assertEqual(
            output.splitlines()[-1], "🖐 Configuration left unchanged"
        )

    def test_save_with_token(self) -> None:
        """Test the save command with a token already present."""
        cfg = self._config()
        cfg.set_api_token("12345678")
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with mock.patch(
            "debusine.client.setup.DebusineClientConfig.save"
        ) as save:
            self.assertTrue(editor.onecmd("save"))
        save.assert_called()
        self.assertEqual(
            console.output.getvalue().splitlines(), [FEEDBACK_CONFIG_SAVED]
        )

    def test_save_fetch_token_confirm(self) -> None:
        """Test the save command with a token, user confirmed."""
        cfg = self._config()

        def _mock_fetch_token(
            console: BufferConsole,  # noqa: ARG001, U100
        ) -> bool:
            cfg.set_api_token("12345678")
            return True

        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with (
            mock.patch.object(cfg, "save") as save,
            mock.patch.object(
                cfg, "fetch_token", side_effect=_mock_fetch_token
            ) as fetch_token,
        ):
            self.assertTrue(editor.onecmd("save"))
        fetch_token.assert_called_with(console)
        save.assert_called()
        self.assertEqual(cfg.server.api_token, "12345678")
        self.assertEqual(
            console.output.getvalue().splitlines(), [FEEDBACK_CONFIG_SAVED]
        )

    def test_save_fetch_token_cancel(self) -> None:
        """Test the save command with a token, user cancelled."""
        cfg = self._config()

        def _mock_fetch_token(
            console: BufferConsole,  # noqa: ARG001, U100
        ) -> bool:
            cfg.set_api_token(None)
            return True

        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with (
            mock.patch.object(cfg, "save") as save,
            mock.patch.object(
                cfg, "fetch_token", side_effect=_mock_fetch_token
            ) as fetch_token,
        ):
            self.assertTrue(editor.onecmd("save"))
        fetch_token.assert_called_with(console)
        save.assert_called()
        self.assertIsNone(cfg.server.api_token)
        self.assertEqual(
            console.output.getvalue().splitlines(), [FEEDBACK_CONFIG_SAVED]
        )

    def test_save_retry(self) -> None:
        """Test the save command with a token."""
        cfg = self._config()

        call_count = 0

        def _mock_fetch_token(
            console: BufferConsole,  # noqa: ARG001, U100
        ) -> bool:
            nonlocal call_count
            try:
                match call_count:
                    case 0:
                        return False
                    case _:
                        cfg.set_api_token("12345678")
                        return True
            finally:
                call_count += 1

        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with (
            mock.patch.object(cfg, "save") as save,
            mock.patch.object(
                cfg, "fetch_token", side_effect=_mock_fetch_token
            ) as fetch_token,
            mock.patch(
                "debusine.client.setup.Confirm.ask", return_value=True
            ) as ask,
        ):
            self.assertTrue(editor.onecmd("save"))
        fetch_token.assert_called_with(console)
        ask.assert_called_with("Fetch failed: retry?", console=console)
        save.assert_called()
        self.assertEqual(cfg.server.api_token, "12345678")
        self.assertEqual(
            console.output.getvalue().splitlines(), [FEEDBACK_CONFIG_SAVED]
        )

    def test_save_abort(self) -> None:
        """Test the save command with a token."""
        cfg = self._config()
        console = BufferConsole()
        editor = ServerConfigEditor(cfg, console=console)
        with (
            mock.patch.object(cfg, "save") as save,
            mock.patch.object(
                cfg, "fetch_token", return_value=False
            ) as fetch_token,
            mock.patch(
                "debusine.client.setup.Confirm.ask", return_value=False
            ) as ask,
        ):
            self.assertTrue(editor.onecmd("save"))
        fetch_token.assert_called_with(console)
        ask.assert_called_with("Fetch failed: retry?", console=console)
        save.assert_called()
        self.assertIsNone(cfg.server.api_token)
        self.assertEqual(
            console.output.getvalue().splitlines(), [FEEDBACK_CONFIG_SAVED]
        )


class SetupServerTests(TestCase):
    """Tests for :py:func:`setup_server`."""

    def test_no_args(self) -> None:
        """Invoke with no arguments."""
        workdir = self.create_temporary_directory()
        console = BufferConsole()

        def _select_server(self: ServerSelector) -> None:
            self.selected = ServerInfo.from_string("name", desc="desc")

        with (
            mock.patch(
                "debusine.client.setup.ServerSelector.cmdloop",
                autospec=True,
                side_effect=_select_server,
            ) as selector_cmdloop,
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=workdir / "config.ini", console=console
            )

        selector_cmdloop.assert_called()
        selector = selector_cmdloop.call_args.args[0]
        self.assertEqual(selector.entries, list(KNOWN_SERVERS.values()))
        editor_cmdloop.assert_called()
        editor = editor_cmdloop.call_args.args[0]
        server = editor.config.server
        self.assertEqual(server.name, "name")
        self.assertEqual(server.scope, "debusine")

    def test_no_server_selected(self) -> None:
        """Invoke with no arguments."""
        workdir = self.create_temporary_directory()
        console = BufferConsole()

        with (
            mock.patch(
                "debusine.client.setup.ServerSelector.cmdloop",
            ) as selector_cmdloop,
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=workdir / "config.ini", console=console
            )

        selector_cmdloop.assert_called()
        editor_cmdloop.assert_not_called()

    def test_server_provided(self) -> None:
        """Invoke with server set."""
        workdir = self.create_temporary_directory()
        console = BufferConsole()

        with (
            mock.patch(
                "debusine.client.setup.ServerSelector.cmdloop",
            ) as selector_cmdloop,
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=workdir / "config.ini",
                server="name",
                console=console,
            )

        selector_cmdloop.assert_not_called()
        editor_cmdloop.assert_called()
        editor = editor_cmdloop.call_args.args[0]
        server = editor.config.server
        self.assertEqual(server.name, "name")
        self.assertEqual(server.scope, "debusine")
        self.assertEqual(editor.config.default_server, "name")

    def test_scope_provided(self) -> None:
        """Invoke with scope set."""
        workdir = self.create_temporary_directory()
        console = BufferConsole()

        def _select_server(self: ServerSelector) -> None:
            self.selected = ServerInfo.from_string("name", desc="desc")

        with (
            mock.patch(
                "debusine.client.setup.ServerSelector.cmdloop",
                autospec=True,
                side_effect=_select_server,
            ) as selector_cmdloop,
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=workdir / "config.ini",
                scope="scope",
                console=console,
            )

        selector_cmdloop.assert_called()
        selector = selector_cmdloop.call_args.args[0]
        self.assertEqual(selector.entries, list(KNOWN_SERVERS.values()))
        editor_cmdloop.assert_called()
        editor = editor_cmdloop.call_args.args[0]
        server = editor.config.server
        self.assertEqual(server.name, "name")
        self.assertEqual(server.scope, "scope")

    def test_existing_config(self) -> None:
        """Invoke with an existing configuration."""
        workdir = self.create_temporary_directory()
        config_path = workdir / "config.ini"
        config_path.write_text(
            textwrap.dedent(
                """
                [General]
                default-server=localhost
                [server:localhost]
                scope = debusine
                api-url = http://localhost:8000/api/
                [server:debian]
                scope = debian
                api-url = http://debusine.debian.net/api/
                [server:example]
                """
            )
        )
        console = BufferConsole()

        def _select_server(self: ServerSelector) -> None:
            self.selected = KNOWN_SERVERS["debian"]

        with (
            mock.patch(
                "debusine.client.setup.ServerSelector.cmdloop",
                autospec=True,
                side_effect=_select_server,
            ) as selector_cmdloop,
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=config_path,
                console=console,
            )

        selector_cmdloop.assert_called()
        selector = selector_cmdloop.call_args.args[0]
        self.assertEqual(
            [x.name for x in selector.entries],
            ["localhost", "debian", "example", "freexian"],
        )
        editor_cmdloop.assert_called()
        editor = editor_cmdloop.call_args.args[0]
        server = editor.config.server
        self.assertEqual(server.name, "debian")
        self.assertEqual(server.scope, "debian")
        self.assertEqual(editor.config.default_server, "localhost")

    def test_existing_config_override_scope(self) -> None:
        """Invoke with an existing configuration, setting a scope."""
        workdir = self.create_temporary_directory()
        config_path = workdir / "config.ini"
        config_path.write_text(
            textwrap.dedent(
                """
                [server:localhost]
                scope = debusine
                api-url = http://localhost:8000/api/
                """
            )
        )
        console = BufferConsole()

        with (
            mock.patch(
                "debusine.client.setup.ServerConfigEditor.cmdloop",
                autospec=True,
            ) as editor_cmdloop,
        ):
            setup_server(
                config_file_path=config_path,
                server="localhost",
                scope="debian",
                console=console,
            )

        editor_cmdloop.assert_called()
        editor = editor_cmdloop.call_args.args[0]
        server = editor.config.server
        self.assertEqual(server.name, "localhost")
        self.assertEqual(server.scope, "debian")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the signing models."""

import random

import pydantic
from django.test import TestCase
from nacl.public import PrivateKey

from debusine.signing.models import (
    AvailableKeyFileSystem,
    AvailableKeyPKCS11,
    ProtectedKeyError,
    ProtectedKeyNaCl,
    ProtectedKeyPKCS11Static,
)


class ProtectedKeyNaClTests(TestCase):
    """Unit tests for `ProtectedKeyNaCl`."""

    def test_encrypt_error(self) -> None:
        """Encryption errors raise ProtectedKeyError."""
        # This doesn't fail in normal use unless libsodium is broken.  Force
        # it to fail using a type error.
        public_key = PrivateKey.generate().public_key
        with self.assertRaisesRegex(
            ProtectedKeyError, "input message must be bytes"
        ):
            ProtectedKeyNaCl.encrypt(
                public_key,
                "should be bytes",  # type: ignore[arg-type]
            )

    def test_decrypt_wrong_key(self) -> None:
        """Decrypting with the wrong key raises ProtectedKeyError."""
        private_keys = [PrivateKey.generate() for _ in range(2)]
        protected = ProtectedKeyNaCl.encrypt(
            private_keys[0].public_key, b"data"
        )
        with self.assertRaisesRegex(
            ProtectedKeyError,
            "Key not encrypted using any of the given private keys",
        ):
            protected.decrypt([private_keys[1]])

    def test_decrypt_error(self) -> None:
        """Other decryption errors raise ProtectedKeyError."""
        private_key = PrivateKey.generate()
        protected = ProtectedKeyNaCl.create(
            public_key=bytes(private_key.public_key), encrypted=b"nonsense" * 8
        )
        with self.assertRaisesRegex(
            ProtectedKeyError, "An error occurred trying to decrypt the message"
        ):
            protected.decrypt([private_key])

    def test_encrypt_decrypt(self) -> None:
        """Encrypting and decrypting some data works."""
        private_key = PrivateKey.generate()
        protected = ProtectedKeyNaCl.encrypt(private_key.public_key, b"data")
        self.assertEqual(protected.decrypt([private_key]), b"data")

    def test_encrypt_decrypt_rotated(self) -> None:
        """The key-encryption key can be rotated."""
        private_keys = [PrivateKey.generate() for _ in range(2)]
        protected = ProtectedKeyNaCl.encrypt(
            private_keys[1].public_key, b"data"
        )
        self.assertEqual(protected.decrypt(private_keys), b"data")

    def test_available(self) -> None:
        """The key can be made temporarily available."""
        storage_private_key = PrivateKey.generate()
        private_key = random.randbytes(64)
        public_key = random.randbytes(64)
        protected = ProtectedKeyNaCl.encrypt(
            storage_private_key.public_key, private_key
        )

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            protected.available(public_key) as available,
        ):
            assert isinstance(available, AvailableKeyFileSystem)
            temp_dir = available.temp_dir
            key_path = temp_dir / "tmp.key"
            certificate_path = temp_dir / "tmp.crt"
            self.assertEqual(key_path.read_bytes(), private_key)
            self.assertEqual(certificate_path.read_bytes(), public_key)
            self.assertEqual(
                available.sbsign_args,
                ["--key", str(key_path), "--cert", str(certificate_path)],
            )

        # After the context manager exits, the key and certificate are no
        # longer on the file system.
        self.assertFalse(temp_dir.exists())


class ProtectedKeyPKCS11StaticTests(TestCase):
    """Unit tests for `ProtectedKeyPKCS11Static`."""

    def test_bad_pkcs11_uri(self) -> None:
        """The PKCS#11 URI must use the "pkcs11" scheme."""
        with self.assertRaisesRegex(
            pydantic.ValidationError, "pkcs11_uri must start with 'pkcs11:'"
        ):
            ProtectedKeyPKCS11Static.create(pkcs11_uri="https://example.com/")

    def test_available(self) -> None:
        """The key can be made temporarily available."""
        public_key = random.randbytes(64)
        protected = ProtectedKeyPKCS11Static.create(pkcs11_uri="pkcs11:id=1")

        with protected.available(public_key) as available:
            assert isinstance(available, AvailableKeyPKCS11)
            temp_dir = available.temp_dir
            certificate_path = temp_dir / "tmp.crt"
            self.assertEqual(certificate_path.read_bytes(), public_key)
            self.assertEqual(
                available.sbsign_args,
                [
                    "--engine",
                    "pkcs11",
                    "--key",
                    "pkcs11:id=1",
                    "--cert",
                    str(certificate_path),
                ],
            )

        # After the context manager exits, the certificate is no longer on
        # the file system.
        self.assertFalse(temp_dir.exists())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: collections."""

import logging
import re
from typing import Any, override

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.context import context
from debusine.db.models import Collection, CollectionItem, CollectionTokenGrant
from debusine.db.models.collections import (
    CollectionQuerySet,
    CollectionTokenGrantQuerySet,
)
from debusine.db.models.permissions import anonymous_access
from debusine.server.collections import DebusineTaskConfigurationManager
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    BaseSerializerBase,
    CollectionSerializer,
    CollectionTokenGrantSerializer,
    TaskConfigurationCollectionContents,
    TaskConfigurationCollectionContentsSerializer,
    TaskConfigurationCollectionUpdateResultsSerializer,
)
from debusine.server.views.base import BaseAPIView, GenericViewSetBase
from debusine.server.views.rest import IsUserTokenAuthenticated

logger = logging.getLogger(__name__)


class TaskConfigurationCollectionView(BaseAPIView):
    """View used to fetch and update a task configuration collection."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]

    def _get_collection(self, workspace: str, name: str) -> Collection:
        """Lookup the task configuration collection."""
        self.set_current_workspace(workspace)
        try:
            collection = context.require_workspace().get_collection(
                pc=context.pc,
                category=CollectionCategory.TASK_CONFIGURATION,
                name=name,
            )
        except Collection.DoesNotExist:
            raise DebusineAPIException(
                title="Collection not found",
                detail=(
                    f"Task configuration collection {name!r} does not exist"
                    f" in workspace {workspace!r}"
                ),
                status_code=status.HTTP_404_NOT_FOUND,
            )

        self.enforce(collection.can_display)
        return collection

    def get(
        self,
        request: Request,  # noqa: ARG002, U100
        workspace: str,
        name: str,
    ) -> Response:
        """Retrieve the contents of the collection."""
        collection = self._get_collection(workspace, name)

        items = [
            DebusineTaskConfiguration.model_validate(child.data)
            for child in collection.child_items.active().filter(
                child_type=CollectionItem.Types.BARE
            )
        ]

        contents = TaskConfigurationCollectionContents(
            collection=collection, items=items
        )
        return Response(
            TaskConfigurationCollectionContentsSerializer(contents).data,
            status=status.HTTP_200_OK,
        )

    @staticmethod
    def _diff_items(
        collection: Collection, items: list[DebusineTaskConfiguration]
    ) -> tuple[list[CollectionItem], list[DebusineTaskConfiguration], int]:
        """
        Compare the collection contents against a new list of items.

        :returns: a tuple (removed, added, unchanged), where:
         * removed lists the CollectionItem entries to remove
         * added lists the DebusineTaskConfiguration items to add
         * unchanged is the count of unchanged entries
        """
        # Updates are not supported: something as small as changing a typo in
        # a comment deletes the old item and re-adds it: that way we do not
        # rewrite history

        # Load the existing items from the database
        old: dict[str, CollectionItem] = {}
        for child in collection.child_items.active().filter(
            child_type=CollectionItem.Types.BARE
        ):
            item = DebusineTaskConfiguration.model_validate(child.data)
            old[item.model_dump_json(exclude_unset=True)] = child

        # Index new items for comparison
        new: dict[str, DebusineTaskConfiguration] = {
            item.model_dump_json(exclude_unset=True): item for item in items
        }

        return (
            # Removed
            [old[k] for k in old.keys() - new.keys()],
            # Added
            [new[k] for k in new.keys() - old.keys()],
            # Unchanged
            len(old.keys() & new.keys()),
        )

    def _validate_git_commit(self, value: Any, name: str) -> str:
        if not isinstance(value, str):
            raise DebusineAPIException(
                title="Invalid git commit",
                detail=f"{name} is not a string",
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        if not re.match(r"^[0-9a-f]{40,64}$", value):
            raise DebusineAPIException(
                title="Invalid git commit",
                detail=f"{name} is malformed",
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        return value

    def post(self, request: Request, workspace: str, name: str) -> Response:
        """Update the contents of the collection."""
        collection = self._get_collection(workspace, name)
        self.enforce(collection.workspace.can_edit_task_configuration)
        dry_run = request.data.pop("dry_run", False)
        serializer = TaskConfigurationCollectionContentsSerializer(
            data=request.data
        )
        serializer.is_valid(raise_exception=True)
        submitted_collection = serializer.validated_data["collection"]

        if submitted_collection["id"] != collection.pk:
            raise DebusineAPIException(
                title="Posted collection mismatch",
                detail=(
                    f"Data posted for collection {submitted_collection['id']}"
                    f" to update collection {collection.pk}"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        if submitted_collection["name"] != collection.name:
            raise DebusineAPIException(
                title="Posted collection mismatch",
                detail=(
                    "Data posted for collection"
                    f" {submitted_collection['name']!r}"
                    f" to update collection {collection.name!r}"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # Update git commit in collection data
        if (
            git_commit := submitted_collection["data"].get("git_commit")
        ) is not None:
            git_commit = self._validate_git_commit(git_commit, "git_commit")
            collection.data["git_commit"] = git_commit
        else:
            collection.data.pop("git_commit", None)
        collection.save()

        # We do not update existing items, as that would rewrite collection
        # history. If an item has been changed, we mark the old version as
        # deleted and create a new version
        removed, added, unchanged = self._diff_items(
            collection, serializer.validated_data["items"]
        )
        removed_names = {item.name for item in removed}

        # Update collection
        manager = DebusineTaskConfigurationManager(collection=collection)
        user = context.user
        assert user.is_authenticated
        if not dry_run:
            for removed_item in removed:
                manager.remove_item(removed_item, user=user)

        # Count updates as the number of added names that are also in removed
        updated: int = 0
        for added_item in added:
            if not dry_run:
                name = manager.add_bare_data(
                    BareDataCategory.TASK_CONFIGURATION,
                    user=user,
                    data=added_item,
                ).name
            else:
                name = added_item.name()
            if name in removed_names:
                updated += 1

        return Response(
            TaskConfigurationCollectionUpdateResultsSerializer(
                {
                    "added": len(added) - updated,
                    "removed": len(removed) - updated,
                    "updated": updated,
                    "unchanged": unchanged,
                }
            ).data,
            status=status.HTTP_200_OK,
        )


class CollectionViewSet(
    BaseAPIView,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    GenericViewSetBase[Collection],
):
    """Provide access to collections."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]
    serializer_class = CollectionSerializer

    @override
    def initial(self, request: Request, *args: Any, **kwargs: Any) -> None:
        super().initial(request, *args, **kwargs)
        self.set_current_workspace(kwargs["workspace"])

    @override
    def get_queryset(self) -> CollectionQuerySet[Any]:
        """Filter accessible collections."""
        self.enforce(anonymous_access)  # filtered to can_display
        queryset = (
            Collection.objects.can_display(context.pc)
            .filter(workspace=context.require_workspace())
            .order_by("category", "name")
        )
        if self.action == "list":
            queryset = queryset.exclude(
                category=CollectionCategory.WORKFLOW_INTERNAL
            )
        return queryset

    @override
    def get_serializer_context(self) -> dict[str, Any]:
        context = super().get_serializer_context()
        context['with_stats'] = self.request.GET.get("stats") == "1"
        return context

    @action(
        detail=False,
        url_path=r"(?P<category>[^/]+)/(?P<name>[^/]+)",
        url_name="lookup",
    )
    def lookup(
        self,
        request: Request,  # noqa: ARG002, U100
        workspace: str,  # noqa: ARG002, U100
        category: str,
        name: str,
    ) -> Response:
        """Look up a collection by name."""
        collection = get_object_or_404(
            self.get_queryset(), name=name, category=category
        )
        serializer = self.get_serializer(collection)
        return Response(serializer.data)

    def set_name(self, collection: Collection, name: str) -> bool:
        """
        Rename the collection.

        :return: True if the name has been changed
        """
        if name == collection.name:
            return False

        if (
            collection.category in SINGLETON_COLLECTION_CATEGORIES
            and name != "_"
        ):
            raise DebusineAPIException(
                title="Invalid name for singleton collection",
                detail=(
                    f"'{collection.category}' singleton collections"
                    " must be named '_'"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        if Collection.objects.filter(
            workspace=collection.workspace,
            category=collection.category,
            name=name,
        ).exists():
            raise DebusineAPIException(
                title="New name conflicts with an existing collection",
                detail=(
                    "A collection with category"
                    f" '{collection.category}' called {name!r}"
                    " already exists"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        collection.name = name
        return True

    def set_full_history_retention_period(
        self, collection: Collection, value: int | None
    ) -> bool:
        """
        Set the full history retention period.

        :return: True if the value has been changed
        """
        if value == collection.full_history_retention_period:
            return False
        collection.full_history_retention_period = value
        return True

    def set_metadata_only_retention_period(
        self, collection: Collection, value: int | None
    ) -> bool:
        """
        Set the full history retention period.

        :return: True if the value has been changed
        """
        if value == collection.metadata_only_retention_period:
            return False
        collection.metadata_only_retention_period = value
        return True

    def set_data(self, collection: Collection, value: dict[str, Any]) -> bool:
        """
        Set the collection data.

        :return: True if the value has been changed
        """
        if value == collection.data:
            return False
        # TODO: The key-value data should be validated against the
        # collection's pydantic data model, once we have it.
        # See https://salsa.debian.org/freexian-team/debusine/-/issues/923
        collection.data = value
        return True

    @override
    def perform_create(self, serializer: BaseSerializerBase[Any]) -> None:
        self.enforce(context.require_workspace().can_create_collections)
        try:
            serializer.save(workspace=context.require_workspace())
        except IntegrityError as e:
            detail = None
            message = str(e)
            if "db_collection_unique_name_category_workspace" in message:
                detail = "A collection with the same name already exists"
            raise DebusineAPIException(
                title="Cannot create collection", detail=detail
            )

    @override
    def perform_update(self, serializer: BaseSerializerBase[Any]) -> None:
        assert isinstance(serializer, CollectionSerializer)
        collection = serializer.instance
        assert isinstance(collection, Collection)
        self.enforce(collection.can_configure)

        changed: bool = False
        if "category" in serializer.validated_data:
            if serializer.validated_data["category"] != collection.category:
                raise DebusineAPIException(
                    title="The category of a collection cannot be changed",
                    status_code=status.HTTP_400_BAD_REQUEST,
                )
        if "name" in serializer.validated_data:
            changed = (
                self.set_name(collection, serializer.validated_data["name"])
                or changed
            )
        if "full_history_retention_period" in serializer.validated_data:
            changed = (
                self.set_full_history_retention_period(
                    collection,
                    serializer.validated_data["full_history_retention_period"],
                )
                or changed
            )
        if "metadata_only_retention_period" in serializer.validated_data:
            changed = (
                self.set_metadata_only_retention_period(
                    collection,
                    serializer.validated_data["metadata_only_retention_period"],
                )
                or changed
            )
        if "data" in serializer.validated_data:
            changed = (
                self.set_data(collection, serializer.validated_data["data"])
                or changed
            )

        if changed:
            collection.save()

    @override
    def perform_destroy(self, instance: Collection) -> None:
        self.enforce(instance.can_delete)
        instance.delete_collection()


class CollectionTokenGrantViewSet(
    BaseAPIView,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
    GenericViewSetBase[CollectionTokenGrant],
):
    """Provide access to tokens granted roles on collections."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]
    serializer_class = CollectionTokenGrantSerializer

    collection: Collection

    @override
    def initial(self, request: Request, *args: Any, **kwargs: Any) -> None:
        """Return the initial request object."""
        super().initial(request, *args, **kwargs)
        self.set_current_workspace(kwargs["workspace"])
        workspace = context.require_workspace()
        self.enforce(workspace.can_display)
        try:
            self.collection = workspace.get_collection(
                pc=context.pc, category=kwargs["ccat"], name=kwargs["cname"]
            )
        except Collection.DoesNotExist:
            raise DebusineAPIException(
                title="Collection not found",
                detail=(
                    f"Collection '{kwargs['cname']}@{kwargs['ccat']}' does not"
                    f" exist in workspace {workspace}"
                ),
                status_code=status.HTTP_404_NOT_FOUND,
            )
        self.enforce(self.collection.can_display)

    @classmethod
    def get_lookup_regex(cls, lookup_prefix: str = "") -> str:
        """Get the portion of the URL regex used to match a single grant."""
        return (
            f"(?P<{lookup_prefix}token>[^/]+)/(?P<{lookup_prefix}role>[^/.]+)"
        )

    @override
    def get_queryset(self) -> CollectionTokenGrantQuerySet[Any]:
        """Filter accessible and relevant tokens."""
        return self.collection.token_grants.order_by(
            Coalesce("token__user__username", "token__external_username"),
            "token__hash",
            "role",
            "id",
        )

    @override
    def get_object(self) -> CollectionTokenGrant:
        queryset = self.filter_queryset(self.get_queryset())
        obj = get_object_or_404(
            queryset, token__hash=self.kwargs["token"], role=self.kwargs["role"]
        )
        assert isinstance(obj, CollectionTokenGrant)
        self.check_object_permissions(self.request, obj)
        return obj

    @override
    def perform_create(
        self, serializer: BaseSerializerBase[CollectionTokenGrant]
    ) -> None:
        self.enforce(self.collection.can_configure)
        try:
            serializer.save(collection=self.collection, created_by=context.user)
        except ValidationError as e:
            raise DebusineAPIException(
                title="Cannot grant collection token",
                validation_errors=e.message_dict,
            )

    @override
    def perform_destroy(self, instance: CollectionTokenGrant) -> None:
        self.enforce(self.collection.can_configure)
        super().perform_destroy(instance)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine checks using Django checks framework."""

import getpass
import os
import pathlib
from collections.abc import Iterable, Sequence
from functools import partial
from typing import Any

from django.apps.config import AppConfig
from django.conf import settings
from django.core.checks import Error, register

from debusine.django.checks import all_migrations_applied, is_migration_command


def _check_directories_owners(
    setting_names: Iterable[str], *, missing_ok: bool
) -> list[Error]:
    current_user = getpass.getuser()
    acceptable_owners = {'root', current_user}

    errors = []

    for setting_name in setting_names:
        directory = pathlib.Path(getattr(settings, setting_name))

        if not directory.exists() and missing_ok:
            continue
        elif not directory.exists():
            errors.append(Error(f"{directory} must exist"))
        elif directory.owner() not in acceptable_owners:
            errors.append(
                Error(
                    f"{directory} owner ({directory.owner()}) must match "
                    f"current user ({current_user})"
                )
            )

    return errors


@register()
def directories_owners_check(
    app_configs: Sequence[AppConfig] | None,  # noqa: ARG001, U100
    **kwargs: Any,
) -> list[Error]:
    """Check optional directories, if they exist, have the correct owner."""
    if is_migration_command():
        return []

    return _check_directories_owners(
        [
            "DEBUSINE_DATA_PATH",
            "STATIC_ROOT",
            "MEDIA_ROOT",
            "DEBUSINE_CACHE_DIRECTORY",
            "DEBUSINE_TEMPLATE_DIRECTORY",
            "DEBUSINE_UPLOAD_DIRECTORY",
            "DEBUSINE_STORE_DIRECTORY",
            "DEBUSINE_TOKENS_DIRECTORY",
        ],
        missing_ok=True,
    )


@register()
def disable_automatic_scheduling_false_in_production(
    app_configs: Sequence[AppConfig] | None,  # noqa: ARG001, U100
    **kwargs: Any,
) -> list[Error]:
    """Check DISABLE_AUTOMATIC_SCHEDULING is enabled only in TEST_MODE."""
    if getattr(settings, "DISABLE_AUTOMATIC_SCHEDULING", False) and not getattr(
        settings, "TEST_MODE", False
    ):
        return [
            Error(
                "DISABLE_AUTOMATIC_SCHEDULING can only be enabled in TEST_MODE",
                hint="Remove DISABLE_AUTOMATIC_SCHEDULING=True "
                "from debusine-server settings",
            )
        ]

    return []


register(partial(all_migrations_applied, "debusine-admin"))


@register()
def local_file_store_directories_writable(
    app_configs: Sequence[AppConfig] | None,  # noqa: ARG001, U100
    **kwargs: Any,
) -> list[Error]:
    """
    Check that each local file store base_directory exist.

    Avoid checking it if a migration command is being run.
    """
    if is_migration_command():
        return []

    from debusine.db.models import FileStore
    from debusine.server.file_backend.local import LocalFileBackend

    errors = []

    for file_store in FileStore.objects.filter(
        backend=FileStore.BackendChoices.LOCAL
    ).order_by("name"):
        file_store_backend = file_store.get_backend_object()
        assert isinstance(file_store_backend, LocalFileBackend)
        base_directory = file_store_backend.base_directory()
        if not os.access(base_directory, os.W_OK):
            errors.append(
                Error(
                    f'FileStore "{file_store.name}" has a non-existent '
                    f'or non-writable base_directory: "{base_directory}"',
                    hint="Create the directory, change the permissions "
                    "or change the FileStore configuration",
                )
            )

    return errors



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests DebusineAsyncHttpClient."""

import asyncio.log
import itertools
import logging
from typing import override

from debusine.test import TestCase
from debusine.utils import get_user_agent
from debusine.worker.config import ConfigHandler
from debusine.worker.debusine_async_http_client import DebusineAsyncHttpClient
from debusine.worker.tests import server


class TestDebusineAsyncHttpClient(server.DebusineAioHTTPTestCase, TestCase):
    """Tests for DebusineAsyncHttpClient."""

    @override
    async def setUpAsync(self) -> None:
        # asyncio/base_events.py print "<some information> is serving" in the
        # debug log. Unit tests can be executed with environment variables:
        # PYTHONASYNCDEBUG=1 PYTHONDEBUG=1 (which has been useful in the past)
        # but the "<some information> is serving" message can be discarded.
        logger = logging.getLogger(asyncio.log.logger.name)

        logger.addFilter(lambda record: not record.msg.endswith("is serving"))

        await super().setUpAsync()
        self.debusine_url = str(self.server.make_url(''))

        self.config = self.build_basic_config()

        self.debusine_async_http_client = DebusineAsyncHttpClient(self.config)

    @override
    async def tearDownAsync(self) -> None:
        await super().tearDownAsync()
        await self.debusine_async_http_client.close()

    def build_basic_config(
        self, overwrite_url: str | None = None
    ) -> ConfigHandler:
        """Build a basic configuration with log-file = /dev/null."""
        url = overwrite_url or self.debusine_url

        config_temp_directory = self.create_temp_config_directory(
            {'General': {'api-url': url}}
        )

        config = ConfigHandler(directories=[config_temp_directory.as_posix()])
        config['General']['log-file'] = '/dev/null'

        return config

    def setup_valid_token(self, include_token: bool) -> None:
        """Set a valid token in the test server and client configuration."""
        self.config = self.build_basic_config()

        if include_token:
            token = '6c931875627131b5135b7de3371c44'
            self.server_config.registered_token = token
            self.config.write_token(token)
        else:
            self.server_config.registered_token = None

        self.debusine_async_http_client._config = self.config

    async def test_make_http_request_raise_path_must_be_absolute(self) -> None:
        """Worker._make_http_request raise exception for relative paths."""
        with self.assertRaisesRegex(ValueError, 'Path must be absolute'):
            self.debusine_async_http_client._make_http_request(
                'PUT', 'api/endpoint'
            )

    async def test_make_http_request_raise_invalid_method(self) -> None:
        """Worker._make_http_request raise exception for unknown HTTP method."""
        with self.assertRaisesRegex(
            ValueError, '^Method must be one of: PUT, POST, GET$'
        ):
            self.debusine_async_http_client._make_http_request(
                'connect', '/api/endpoint'
            )

    async def make_http_request(
        self, method: str, path: str, token: str | None = None
    ) -> dict[str, str] | None:
        """
        Make HTTP request, return data sent (or None if no data was sent).

        :param method: HTTP method (GET, POST...)
        :param path: path for the request.
        """
        if method != 'GET':
            data = {'key': 'data'}
        else:
            data = None

        await self.debusine_async_http_client._make_http_request(
            method, path, json=data, token=token
        )

        return data

    async def assert_http_request(
        self,
        method: str,
        path: str,
        include_token: bool,
        data: dict[str, str] | None,
    ) -> None:
        """
        Assert that Worker._make_http_request sends the data and token.

        :param method: HTTP method used for the test.
        :param path: path to match the request.
        :param include_token: includes the token in the headers or not.
        :param data: data to match the request.
        """
        # We've always performed at least one request by this point.
        assert self.server_latest_request is not None
        assert self.server_latest_request.headers is not None

        # Token was included
        self.assertEqual(
            'token' in self.server_latest_request.headers, include_token
        )

        # HTTP method sent is the method received
        self.assertEqual(method, self.server_latest_request.method)

        # Data sent from the client is the data received in the server
        self.assertEqual(self.server_latest_request.json_content, data)

        # Path that the request was made is the path that it was received
        self.assertEqual(self.server_latest_request.path, path)

    async def test_make_http_request_user_agent(self) -> None:
        """Worker._make_http_request sends the correct User-Agent."""
        self.setup_valid_token(False)
        await self.make_http_request('GET', '/collect_request_information/')

        assert self.server_latest_request is not None
        assert self.server_latest_request.headers is not None
        self.assertEqual(
            self.server_latest_request.headers.get("User-Agent"),
            get_user_agent(),
        )

    async def test_make_http_request(self) -> None:
        """Worker._make_http_request makes the request without token."""
        path = '/collect_request_information/'

        for method, include_token in itertools.product(
            ["PUT", "POST", "GET"], [True, False]
        ):
            with self.subTest(method=method):
                self.setup_valid_token(include_token)
                data = await self.make_http_request(method, path)
                await self.assert_http_request(
                    method, path, include_token=include_token, data=data
                )

    async def test_make_http_request_explicit_token(self) -> None:
        """_make_http_request sends an explicitly-requested token."""
        path = '/collect_request_information/'
        token = 'a84b92e3c3379733bf0d2b25700c4c'

        for method, config_has_token in itertools.product(
            ["PUT", "POST", "GET"], [True, False]
        ):
            with self.subTest(method=method):
                self.setup_valid_token(config_has_token)
                data = await self.make_http_request(method, path, token=token)
                await self.assert_http_request(
                    method, path, include_token=True, data=data
                )

    async def test_make_http_request_get_with_data_raises_value_error(
        self,
    ) -> None:
        """_make_http_request raises ValueError if data passed in HTTP GET."""
        with self.assertRaisesRegex(
            ValueError, '^data not allowed in GET requests$'
        ):
            self.debusine_async_http_client._make_http_request(
                'GET', path='/path/', data={'some_data': 'data'}
            )

    async def test_client_session_uses_configured_timeout(self) -> None:
        """Client session uses http_timeout from config
        (tuple converted to total).
        """
        config = self.build_basic_config()
        config['General']['http-timeout'] = '10,20'  # connect, read

        client = DebusineAsyncHttpClient(config)
        client._ensure_client_session()

        assert client._client_session is not None

        # total = connect + read = 10 + 20
        self.assertEqual(client._client_session.timeout.total, 30)

        await client.close()

    async def test_client_session_uses_int_timeout(self) -> None:
        """Client sessions handles integer timeout from config."""
        config = self.build_basic_config()
        config['General']['http-timeout'] = '25'

        client = DebusineAsyncHttpClient(config)
        client._ensure_client_session()

        assert client._client_session is not None
        self.assertEqual(client._client_session.timeout.total, 25)

        await client.close()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for task input fields."""

import inspect
from collections.abc import Callable
from collections.abc import Collection as AbcCollection
from types import SimpleNamespace
from typing import Any, Self, cast, override
from unittest import mock

from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
    DebusineSigningInput,
    DebusineSigningOutput,
)
from debusine.assets import KeyPurpose
from debusine.client.models import RelationType
from debusine.tasks import BaseExternalTask, TaskConfigError
from debusine.tasks.inputs.inputs import (
    BuildArchitectureInput,
    DataFieldInput,
    DataListFieldInput,
    DebusineFQDNInput,
    EnvironmentInput,
    ExtraRepositoriesInput,
    MultiInput,
    MultiInputList,
    OptionalEnvironmentInput,
    OptionalSingleInput,
    SingleInput,
    SingleInputList,
    SuiteArchiveInput,
    TaskFieldContainer,
    TaskInput,
    UnsignedMultiInput,
    UploadArtifactsInput,
)
from debusine.tasks.inputs.mock_resolver import ConcreteFieldResolver
from debusine.tasks.inputs.resolver import FieldResolver
from debusine.tasks.inputs.serializers import ScalarSerializer
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskInputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
)
from debusine.test import TestCase
from debusine.utils import extract_generic_type_arguments


class InputTests[TI: TaskInput[Any]](TestCase):
    """Test case that can work as a container for task input fields."""

    #: Field class to be tested, filled by __init_subclass__
    field_class: type[TI]
    #: Mock task type to use for self.task
    mock_task_class = BaseExternalTask[Any, Any]

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Fill lookup_type from typing information."""
        super().__init_subclass__(**kwargs)
        [cls.field_class] = extract_generic_type_arguments(cls, InputTests)

    @override
    def setUp(self) -> None:
        super().setUp()
        # self.field is typed here, but set later by test methods
        self.field: TI
        # Task database for resolving fields
        self.field_resolver = ConcreteFieldResolver()
        self.lookup_single_artifact = self.enterContext(
            mock.patch.object(self.field_resolver, "lookup_single_artifact")
        )
        self.lookup_multiple_artifacts = self.enterContext(
            mock.patch.object(self.field_resolver, "lookup_multiple_artifacts")
        )
        self.lookup_environment = self.enterContext(
            mock.patch.object(self.field_resolver, "lookup_environment")
        )
        self.get_server_setting = self.enterContext(
            mock.patch.object(self.field_resolver, "get_server_setting")
        )
        self.get_fallback_architecture = self.enterContext(
            mock.patch.object(self.field_resolver, "get_fallback_architecture")
        )
        self.export_suite_signing_keys = self.enterContext(
            mock.patch.object(self.field_resolver, "export_suite_signing_keys")
        )
        self.find_related_artifacts = self.enterContext(
            mock.patch.object(self.field_resolver, "find_related_artifacts")
        )
        self.get_field = self.enterContext(
            mock.patch.object(self.field_resolver, "get_field")
        )
        # Mock task to use for tests
        self.task = mock.MagicMock(spec=self.mock_task_class)
        self.task.name = "TASKNAME"
        self.task.data = SimpleNamespace()
        # Mock check_category
        self.check_category = self.enterContext(
            mock.patch(
                "debusine.tasks.inputs.inputs.BaseArtifactInput.check_category"
            )
        )

    def set_task_data_field_value(self, value: Any) -> None:
        """Mock self.field.get_field to return the given value."""
        self.get_field.return_value = value

    def resolve(self) -> Any:
        """Call self.field.resolve and return its result."""
        return self.field.resolve(self.field_resolver)

    def make_multi(
        self,
        *artifacts: BaseTaskInputArtifact,
        label: str | None = None,
        lookup: LookupMultiple | None = None,
    ) -> InputArtifactMultiple:
        """Create an InputArtifactMultiple instance."""
        assert hasattr(self.field, "field")
        return InputArtifactMultiple(
            label=label or self.field.field,
            lookup=lookup or self.get_field.return_value,
            artifacts=list(artifacts),
        )

    @staticmethod
    def artifact_info(*, artifact_id: int) -> InputArtifactSingle:
        """Return an InputArtifactSingle with artifact_id and example data."""
        return InputArtifactSingle(
            artifact_id=artifact_id,
            label="Sbuild hello (amd64)",
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="hello",
                srcpkg_version="1.0.0",
                deb_fields={},
                deb_control_files=[],
            ),
            lookup="internal@collections/name:build-amd64",
        )

    def test_has_serializer(self) -> None:
        if not hasattr(self, "field_class") or inspect.isabstract(
            self.field_class
        ):
            return
        if not hasattr(self.field_class, "serializer"):
            self.fail(
                f"{self.field_class} has no serializer defined"
            )  # pragma: no cover


class ConcreteTaskInput(TaskInput[str]):
    """Concrete task input."""

    serializer = ScalarSerializer(str)

    def __init__(self, value: str) -> None:
        """Resolve input field to a string constant."""
        super().__init__()
        self.value = value

    @override
    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[[Self], str]:
        # Reuse an existing method by casting it to the right type
        return cast(Callable[[Self], str], resolver.resolve_suitearchiveinput)


class TaskInputTests(InputTests[ConcreteTaskInput]):
    """Tests for :py:class:`TaskInput`."""

    def test_registered_in_inputs(self) -> None:
        class A(TaskFieldContainer):
            field = ConcreteTaskInput("test")

        self.assertIn("field", A.inputs)

    def test_input_name(self) -> None:
        class A(TaskFieldContainer):
            field = ConcreteTaskInput("test")

        self.assertEqual(A.inputs["field"].name, "field")

    def test_inputs_subclassing(self) -> None:
        class A(TaskFieldContainer):
            field1 = ConcreteTaskInput("one")

        class B(A):
            field2 = ConcreteTaskInput("two")

        self.assertIn("field1", A.inputs)
        self.assertNotIn("field2", A.inputs)
        self.assertIn("field1", B.inputs)
        self.assertIn("field2", B.inputs)

    def test_inputs_subclassing_multiple_inheritance(self) -> None:
        class A(TaskFieldContainer):
            field1 = ConcreteTaskInput("one")

        class B(A):
            field2 = ConcreteTaskInput("two")

        class C(A):
            field3 = ConcreteTaskInput("three")

        class D(C, B):
            field4 = ConcreteTaskInput("four")

        self.assertEqual(list(A.inputs), ["field1"])
        self.assertEqual(list(B.inputs), ["field1", "field2"])
        self.assertEqual(list(C.inputs), ["field1", "field3"])
        self.assertEqual(
            list(D.inputs), ["field1", "field2", "field3", "field4"]
        )

    def test_not_resolved(self) -> None:
        class A(TaskFieldContainer):
            field = ConcreteTaskInput("test")

        a = A()
        with self.assertRaisesRegex(
            TaskConfigError, r"Cannot access unresolved input 'field'"
        ):
            a.field

    def test_default_resolve(self) -> None:
        self.field = ConcreteTaskInput("test")
        resolver_method = self.enterContext(
            mock.patch.object(
                self.field_resolver,
                "resolve_suitearchiveinput",
                return_value=mock.sentinel.result,
            )
        )
        self.assertIs(self.resolve(), mock.sentinel.result)
        resolver_method.assert_called_once_with(self.field)

    def test_get_value(self) -> None:
        class A(TaskFieldContainer):
            field1 = ConcreteTaskInput("one")

        a = A()
        a.__dict__["field1"] = "value"
        self.assertEqual(a.inputs["field1"].get_value(a), "value")

    def test_get_value_not_resolved(self) -> None:
        class A(TaskFieldContainer):
            field1 = ConcreteTaskInput("one")

        a = A()
        with self.assertRaisesRegex(
            TaskConfigError, r"Cannot access unresolved input 'field1'"
        ):
            a.inputs["field1"].get_value(a)

    def test_get_artifact_ids(self) -> None:
        class A(TaskFieldContainer):
            field = ConcreteTaskInput("one")

        a = A()
        a.field = "test"
        self.assertEqual(A.inputs["field"].get_artifact_ids(a.field), [])


class TaskFieldContainerTests(TestCase):
    """Tests for :py:class:`TaskFieldContainer`."""

    @override
    def setUp(self) -> None:
        super().setUp()

        class Container(TaskFieldContainer):
            field = ConcreteTaskInput("test")

        class ChildContainer(TaskFieldContainer):
            child_field = ConcreteTaskInput("test_child")

        self.container = Container()
        self.child_container = ChildContainer()

    def test_field_is_resolved(self) -> None:
        self.assertFalse(self.container.field_is_resolved("field"))
        self.container.field = "foo"
        self.assertTrue(self.container.field_is_resolved("field"))

    def test_iter_all_fields(self) -> None:
        self.assertEqual(
            list(self.container.iter_all_fields()),
            [(self.container, self.container.inputs["field"])],
        )

    def test_iter_all_fields_with_child(self) -> None:
        self.container.child_task_field_container = self.child_container

        self.assertEqual(
            list(self.container.iter_all_fields()),
            [
                (self.container, self.container.inputs["field"]),
                (
                    self.child_container,
                    self.child_container.inputs["child_field"],
                ),
            ],
        )

    def test_save_field_values(self) -> None:
        self.container.field = "field value"

        dynamic_data = BaseDynamicTaskData()
        self.assertNotIn("task_input_fields", dynamic_data.model_fields_set)

        self.container.save_field_values(dynamic_data)

        self.assertEqual(dynamic_data.task_input_fields["field"], "field value")
        self.assertIn("task_input_fields", dynamic_data.model_fields_set)

    def test_save_field_values_with_child(self) -> None:
        self.container.child_task_field_container = self.child_container

        self.container.field = "field value"
        self.child_container.child_field = "child field value"

        dynamic_data = BaseDynamicTaskData()
        self.container.save_field_values(dynamic_data)
        self.assertEqual(dynamic_data.task_input_fields["field"], "field value")
        self.assertEqual(
            dynamic_data.task_input_fields["child_field"], "child field value"
        )

    def test_save_field_values_preserve_existing(self) -> None:
        self.container.field = "field value"

        dynamic_data = BaseDynamicTaskData()
        dynamic_data.task_input_fields["other"] = "other value"

        self.container.save_field_values(dynamic_data)

        self.assertEqual(
            dynamic_data.task_input_fields,
            {
                "field": "field value",
                "other": "other value",
            },
        )

    def test_restore_field_values(self) -> None:
        dynamic_data = BaseDynamicTaskData()
        dynamic_data.task_input_fields["field"] = "field value"
        self.assertFalse(self.container.field_is_resolved("field"))
        self.container.restore_field_values(dynamic_data)
        self.assertTrue(self.container.field_is_resolved("field"))
        self.assertEqual(self.container.field, "field value")

    def test_restore_field_values_with_child(self) -> None:
        self.container.child_task_field_container = self.child_container
        dynamic_data = BaseDynamicTaskData()
        dynamic_data.task_input_fields["field"] = "field value"
        dynamic_data.task_input_fields["child_field"] = "child field value"
        self.assertFalse(self.container.field_is_resolved("field"))
        self.assertFalse(self.child_container.field_is_resolved("child_field"))
        self.container.restore_field_values(dynamic_data)
        self.assertTrue(self.container.field_is_resolved("field"))
        self.assertTrue(self.child_container.field_is_resolved("child_field"))
        self.assertEqual(self.container.field, "field value")
        self.assertEqual(self.child_container.child_field, "child field value")

    def test_restore_field_values_ignores_undefined_fields(self) -> None:
        dynamic_data = BaseDynamicTaskData()
        dynamic_data.task_input_fields["field"] = "field value"
        dynamic_data.task_input_fields["other"] = "other value"
        self.container.restore_field_values(dynamic_data)
        self.assertTrue(self.container.field_is_resolved("field"))
        self.assertEqual(self.container.field, "field value")
        self.assertNotIn("other", self.container.__dict__)


class ConcreteDataFieldInput(DataFieldInput[int, None]):
    """Concrete version of DataFieldInput."""

    serializer = ScalarSerializer(type(None))

    @override
    def resolve(self, field_resolver: FieldResolver) -> None:
        raise NotImplementedError()


class DataFieldInputTests(InputTests[ConcreteDataFieldInput]):
    """Tests for :py:class:`DataFieldInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()

        class Fields(TaskFieldContainer):
            field_default = self.field_class()
            field_test = self.field_class(field="test")

        self.fields = Fields()

    def test_field_default(self) -> None:
        self.assertEqual(
            self.fields.inputs["field_default"].field,  # type: ignore[attr-defined]
            "field_default",
        )

    def test_field_explicit(self) -> None:
        self.assertEqual(
            self.fields.inputs["field_test"].field,  # type: ignore[attr-defined]
            "test",
        )


class ConcreteDataListFieldInput(DataListFieldInput[int, None]):
    """Concrete version of DataListFieldInput."""

    serializer = ScalarSerializer(type(None))

    @override
    def resolve(self, field_resolver: FieldResolver) -> None:
        raise NotImplementedError()


class DataListFieldInputTests(InputTests[ConcreteDataListFieldInput]):
    """Tests for :py:class:`DataListFieldInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()

        class Fields(TaskFieldContainer):
            field_default = self.field_class()
            field_test = self.field_class(field="test")
            field_test_nested = self.field_class(field="test.nested")

        self.fields = Fields()

    def test_field_default(self) -> None:
        self.assertEqual(
            self.fields.inputs["field_default"].field,  # type: ignore[attr-defined]
            "field_default",
        )


class SingleInputTests(InputTests[SingleInput]):
    """Tests for :py:class:`SingleInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_none(self) -> None:
        self.set_task_data_field_value(None)
        with self.assertRaises(AssertionError):
            self.field.resolve(self.field_resolver)
        self.lookup_single_artifact.assert_not_called()
        self.check_category.assert_not_called()

    def test_has_value(self) -> None:
        self.set_task_data_field_value(1)
        self.lookup_single_artifact.return_value = "RESULT"
        self.assertEqual(self.resolve(), "RESULT")
        self.lookup_single_artifact.assert_called_once_with(1, label="field")
        self.check_category.assert_called_with("RESULT")

    def test_get_artifact_ids(self) -> None:
        self.assertEqual(
            self.field.get_artifact_ids(
                self.artifact_info(artifact_id=100),
            ),
            [100],
        )


class OptionalSingleInputTests(InputTests[OptionalSingleInput]):
    """Tests for :py:class:`OptioanlSingleInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_none(self) -> None:
        self.set_task_data_field_value(None)
        self.assertEqual(self.resolve(), None)
        self.lookup_single_artifact.assert_not_called()
        self.check_category.assert_not_called()

    def test_has_value(self) -> None:
        self.set_task_data_field_value("LOOKUP")
        self.lookup_single_artifact.return_value = "RESULT"
        self.assertEqual(self.resolve(), "RESULT")
        self.lookup_single_artifact.assert_called_once_with(
            "LOOKUP", label="field"
        )
        self.check_category.assert_called_with("RESULT")

    def test_get_artifact_ids(self) -> None:
        artifact_info = InputArtifactSingle(
            artifact_id=100,
            label="source",
            category=ArtifactCategory.SOURCE_PACKAGE,
            lookup="",
            data=DebianSourcePackage(
                name="hello",
                version="1.0.0",
                type="dpkg",
                dsc_fields={},
            ),
        )

        self.assertEqual(self.field.get_artifact_ids(artifact_info), [100])
        self.assertEqual(self.field.get_artifact_ids(None), [])


class SingleInputListTests(InputTests[SingleInputList]):
    """Tests for :py:class:`SingleInputList`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_empty_list(self) -> None:
        self.set_task_data_field_value([])
        self.assertEqual(self.resolve(), [])
        self.lookup_single_artifact.assert_not_called()
        self.check_category.assert_not_called()

    def test_multiple_results(self) -> None:
        self.set_task_data_field_value([1, "two"])
        self.lookup_single_artifact.side_effect = ["foo", "bar"]
        self.assertEqual(self.resolve(), ["foo", "bar"])
        self.lookup_single_artifact.assert_has_calls(
            [mock.call(1, label="field"), mock.call("two", label="field")]
        )
        self.check_category.assert_has_calls(
            [mock.call("foo"), mock.call("bar")]
        )

    def test_get_artifact_ids(self) -> None:
        value = [
            self.artifact_info(artifact_id=123),
            self.artifact_info(artifact_id=456),
        ]
        self.assertEqual(self.field.get_artifact_ids(value), [123, 456])


class MultiInputTests(InputTests[MultiInput]):
    """Tests for :py:class:`MultiInputList`."""

    binary = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_empty(self) -> None:
        lookup = LookupMultiple((1,))
        result = InputArtifactMultiple(
            lookup=lookup, label="field", artifacts=[]
        )
        self.set_task_data_field_value(lookup)
        self.lookup_multiple_artifacts.return_value = result
        self.assertEqual(self.resolve(), result)
        self.lookup_multiple_artifacts.assert_called_with(lookup, label="field")
        self.check_category.assert_not_called()

    def test_multiple_results(self) -> None:
        lookup = LookupMultiple((1,))
        result = InputArtifactMultiple(
            lookup=lookup, label="field", artifacts=[self.binary]
        )
        self.set_task_data_field_value(lookup)
        self.lookup_multiple_artifacts.return_value = result
        self.assertEqual(
            self.resolve(), self.lookup_multiple_artifacts.return_value
        )
        self.lookup_multiple_artifacts.assert_called_once_with(
            lookup, label="field"
        )
        self.check_category.assert_not_called()

    def test_check_category(self) -> None:
        self.field.categories = (ArtifactCategory.BINARY_PACKAGE,)
        lookup = LookupMultiple((1,))
        result = InputArtifactMultiple(
            lookup=lookup, label="field", artifacts=[self.binary]
        )
        self.set_task_data_field_value(lookup)
        self.lookup_multiple_artifacts.return_value = result
        self.assertEqual(
            self.resolve(), self.lookup_multiple_artifacts.return_value
        )
        self.lookup_multiple_artifacts.assert_called_once_with(
            lookup, label="field"
        )
        self.check_category.assert_called_once_with(self.binary, "field[0]")

    def test_get_artifact_ids(self) -> None:
        lookup = LookupMultiple((1,))
        value = InputArtifactMultiple(
            lookup=lookup, label="field", artifacts=[self.binary]
        )
        self.assertEqual(self.field.get_artifact_ids(value), [1])


class MultiInputListTests(InputTests[MultiInputList]):
    """Tests for :py:class:`MultiInputList`."""

    binary = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )

    def test_empty_categories(self) -> None:
        self.field = self.field_class()
        self.field.field = "field"

        lookups = [LookupMultiple((1,))]
        self.set_task_data_field_value(lookups)

        results = [
            InputArtifactMultiple(
                lookup=self.get_field.return_value[0],
                label="field",
                artifacts=[self.binary],
            )
        ]
        self.lookup_multiple_artifacts.return_value = results[0]

        self.assertEqual(self.resolve(), results)

        self.lookup_multiple_artifacts.assert_called_with(
            lookups[0], label="field"
        )
        self.check_category.assert_not_called()

    def test_with_categories(self) -> None:
        self.field = MultiInputList(
            field="test", categories=[ArtifactCategory.BINARY_PACKAGE]
        )

        lookups = [
            LookupMultiple((1,)),
            LookupMultiple((2,)),
        ]
        self.set_task_data_field_value(lookups)

        results = [
            InputArtifactMultiple(
                lookup=self.get_field.return_value[0],
                label="field",
                artifacts=[self.binary],
            ),
            InputArtifactMultiple(
                lookup=self.get_field.return_value[1],
                label="field",
                artifacts=[self.binary, self.binary],
            ),
        ]
        self.lookup_multiple_artifacts.side_effect = results
        self.assertEqual(self.resolve(), results)
        self.check_category.assert_has_calls(
            [
                mock.call(results[0].artifacts[0], "test[0][0]"),
                mock.call(results[1].artifacts[0], "test[1][0]"),
                mock.call(results[1].artifacts[1], "test[1][1]"),
            ]
        )

    def test_get_artifact_ids(self) -> None:
        self.field = MultiInputList(
            field="test", categories=[ArtifactCategory.BINARY_PACKAGE]
        )

        value = [
            InputArtifactMultiple(
                label="label1",
                lookup=LookupMultiple(()),
                artifacts=[
                    BaseTaskInputArtifact(artifact_id=100),
                    BaseTaskInputArtifact(artifact_id=101),
                ],
            ),
            InputArtifactMultiple(
                label="label2",
                lookup=LookupMultiple(()),
                artifacts=[BaseTaskInputArtifact(artifact_id=102)],
            ),
            InputArtifactMultiple(
                label="label3",
                lookup=LookupMultiple(()),
                artifacts=[BaseTaskInputArtifact(artifact_id=None)],
            ),
        ]

        self.assertEqual(
            self.field.get_artifact_ids(value),
            [100, 101, 102],
        )


class UploadArtifactsInputTests(InputTests[UploadArtifactsInput]):
    """Tests for :py:class:`UploadArtifactsInput`."""

    binary = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )
    binary1 = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )
    binary2 = BaseTaskInputArtifact(
        artifact_id=2,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned-doc",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "unsigned",
                "Version": "1.0-1",
                "Architecture": "all",
            },
            deb_control_files=[],
        ),
    )
    upload = BaseTaskInputArtifact(
        artifact_id=3,
        category=ArtifactCategory.UPLOAD,
        data=DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "all",
                "Files": [{"name": "unsigned-doc_1.0_all.deb"}],
            },
        ),
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_empty(self) -> None:
        self.set_task_data_field_value([])
        self.assertEqual(self.resolve(), self.make_multi())
        self.find_related_artifacts.assert_not_called()

    def test_no_uploads(self) -> None:
        self.set_task_data_field_value(["LOOKUP"])
        self.lookup_multiple_artifacts.return_value = self.make_multi(
            self.binary
        )
        self.assertEqual(self.resolve(), self.make_multi(self.binary))
        self.lookup_multiple_artifacts.assert_called_once_with(
            ["LOOKUP"], label="field"
        )
        self.find_related_artifacts.assert_not_called()

    def test_expand_uploads(self) -> None:
        self.set_task_data_field_value(["LOOKUP"])
        self.lookup_multiple_artifacts.return_value = self.make_multi(
            self.binary1, self.upload
        )
        self.find_related_artifacts.return_value = self.make_multi(self.binary2)
        result = self.resolve()
        self.assertEqual(
            [x.artifact_id for x in result.artifacts],
            [x.artifact_id for x in (self.binary1, self.binary2)],
        )
        self.find_related_artifacts.assert_called_once_with(
            [self.upload.artifact_id],
            target_category=ArtifactCategory.BINARY_PACKAGE,
            relation_type=RelationType.EXTENDS,
            label="field",
        )


class EnvironmentInputTests(InputTests[EnvironmentInput]):
    """Tests for :py:class:`EnvironmentInput`."""

    def test_resolver_method(self) -> None:
        field = self.field_class()
        self.assertEqual(
            field.get_resolver_method(self.field_resolver),
            self.field_resolver.resolve_environmentinput,
        )

    def test_get_artifact_ids(self) -> None:
        self.field = EnvironmentInput(field="field")
        self.assertEqual(
            self.field.get_artifact_ids(self.artifact_info(artifact_id=123)),
            [123],
        )


class OptionalEnvironmentInputTests(InputTests[OptionalEnvironmentInput]):
    """Tests for :py:class:`OptionalEnvironmentInput`."""

    def test_resolver_method(self) -> None:
        field = self.field_class()
        self.assertEqual(
            field.get_resolver_method(self.field_resolver),
            self.field_resolver.resolve_optionalenvironmentinput,
        )


class SuiteArchiveInputTests(InputTests[SuiteArchiveInput]):
    """Tests for :py:class:`SuiteArchiveInput`."""

    def test_resolver_method(self) -> None:
        field = self.field_class()
        self.assertEqual(
            field.get_resolver_method(self.field_resolver),
            self.field_resolver.resolve_suitearchiveinput,
        )


class ExtraRepositoriesInputTests(InputTests[ExtraRepositoriesInput]):
    """Tests for :py:class:`ExtraRepositoriesInput`."""

    def test_resolver_method(self) -> None:
        field = self.field_class()
        self.assertEqual(
            field.get_resolver_method(self.field_resolver),
            self.field_resolver.resolve_extrarepositoriesinput,
        )


class DebusineFQDNInputTests(InputTests[DebusineFQDNInput]):
    """Tests for :py:class:`DebusineFQDNInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.get_server_setting = self.enterContext(
            mock.patch.object(self.field_resolver, "get_server_setting")
        )

    def test_resolve(self) -> None:
        self.get_server_setting.return_value = "FQDN"
        self.assertEqual(self.resolve(), "FQDN")
        self.get_server_setting.assert_called_once_with("DEBUSINE_FQDN")


class BuildArchitectureInputTests(InputTests[BuildArchitectureInput]):
    """Tests for :py:class:`BuildArchitectureInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_with_value(self) -> None:
        self.set_task_data_field_value("arm64")
        self.assertEqual(self.resolve(), "arm64")
        self.get_fallback_architecture.assert_not_called()

    def test_with_fallback(self) -> None:
        self.set_task_data_field_value(None)
        self.get_fallback_architecture.return_value = (
            mock.sentinel.build_architecture
        )
        self.assertEqual(self.resolve(), mock.sentinel.build_architecture)
        self.get_fallback_architecture.assert_called_once_with()


class UnsignedMultiInputTests(InputTests[UnsignedMultiInput]):
    """Tests for :py:class:`UnsignedMultiInput`."""

    unsigned = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )
    signing_input = BaseTaskInputArtifact(
        artifact_id=2,
        category=ArtifactCategory.SIGNING_INPUT,
        data=DebusineSigningInput(),
    )
    signing_output = BaseTaskInputArtifact(
        artifact_id=3,
        category=ArtifactCategory.SIGNING_OUTPUT,
        data=DebusineSigningOutput(
            purpose=KeyPurpose.UEFI, fingerprint="", results=[]
        ),
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class()
        self.field.field = "field"

    def test_empty(self) -> None:
        self.set_task_data_field_value([])
        self.find_related_artifacts.return_value = self.make_multi()
        self.assertEqual(self.resolve(), self.make_multi())
        self.assertEqual(
            self.find_related_artifacts.mock_calls,
            [
                mock.call(
                    set(),
                    target_category=ArtifactCategory.SIGNING_INPUT,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
                mock.call(
                    set(),
                    target_category=ArtifactCategory.BINARY_PACKAGE,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
            ],
        )

    def test_signing_input(self) -> None:
        self.set_task_data_field_value(["LOOKUP"])
        self.lookup_multiple_artifacts.return_value = self.make_multi(
            self.signing_output
        )

        def find_related_artifacts(
            artifact_ids: AbcCollection[int],
            *,
            target_category: ArtifactCategory,
            relation_type: RelationType = RelationType.RELATES_TO,
            label: str,  # noqa: ARG001
        ) -> InputArtifactMultiple:
            if (
                set(artifact_ids) == {self.signing_output.artifact_id}
                and target_category == ArtifactCategory.SIGNING_INPUT
                and relation_type == RelationType.RELATES_TO
            ):
                return self.make_multi(self.signing_input)
            return self.make_multi()

        self.find_related_artifacts.side_effect = find_related_artifacts

        result = self.resolve()
        self.assertEqual(
            [x.artifact_id for x in result.artifacts],
            [self.signing_input.artifact_id],
        )
        self.assertEqual(
            self.find_related_artifacts.mock_calls,
            [
                mock.call(
                    {self.signing_output.artifact_id},
                    target_category=ArtifactCategory.SIGNING_INPUT,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
                mock.call(
                    {self.signing_input.artifact_id},
                    target_category=ArtifactCategory.BINARY_PACKAGE,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
            ],
        )

    def test_unsigned(self) -> None:
        self.set_task_data_field_value(["LOOKUP"])
        self.lookup_multiple_artifacts.return_value = self.make_multi(
            self.signing_output
        )

        def find_related_artifacts(
            artifact_ids: AbcCollection[int],
            *,
            target_category: ArtifactCategory,
            relation_type: RelationType = RelationType.RELATES_TO,
            label: str,  # noqa: ARG001
        ) -> InputArtifactMultiple:
            if (
                set(artifact_ids) == {self.signing_output.artifact_id}
                and target_category == ArtifactCategory.SIGNING_INPUT
                and relation_type == RelationType.RELATES_TO
            ):
                return self.make_multi(self.signing_input)
            if (
                set(artifact_ids) == {self.signing_input.artifact_id}
                and target_category == ArtifactCategory.BINARY_PACKAGE
                and relation_type == RelationType.RELATES_TO
            ):
                return self.make_multi(self.unsigned)
            raise AssertionError(f"Unexpected artifact_ids: {artifact_ids}")

        self.find_related_artifacts.side_effect = find_related_artifacts
        result = self.resolve()
        self.assertEqual(
            [x.artifact_id for x in result.artifacts],
            sorted(
                [x.artifact_id for x in [self.signing_input, self.unsigned]]
            ),
        )
        self.assertEqual(
            self.find_related_artifacts.mock_calls,
            [
                mock.call(
                    {self.signing_output.artifact_id},
                    target_category=ArtifactCategory.SIGNING_INPUT,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
                mock.call(
                    {self.signing_input.artifact_id},
                    target_category=ArtifactCategory.BINARY_PACKAGE,
                    relation_type=RelationType.RELATES_TO,
                    label='field',
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application's OpenMetrics statistics."""

from collections.abc import Callable, Generator, Mapping
from enum import StrEnum
from typing import Any, override

from celery import current_task
from django.utils.encoding import smart_str
from prometheus_client import Metric
from prometheus_client import exposition as prometheus
from prometheus_client.openmetrics import exposition as openmetrics
from prometheus_client.registry import CollectorRegistry
from rest_framework import renderers, status
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.db.metrics import (
    BadMetricType,
    Collector,
    MetricNotFound,
    emit_metric,
)
from debusine.db.models import Token
from debusine.db.models.permissions import anonymous_access
from debusine.server.exceptions import DebusineAPIException
from debusine.server.open_metrics import DebusineCollector
from debusine.server.prometheus_client.redis_collector import RedisCollector
from debusine.server.serializers import EmitMetricSerializer
from debusine.server.views.base import BaseAPIView
from debusine.server.views.rest import IsWorkerTokenAuthenticated
from debusine.tasks.models import WorkerType


class FilteredCollector:
    """A metric collector that only exposes some metrics from a parent."""

    def __init__(
        self, parent: Collector, accept: Callable[[Metric], bool]
    ) -> None:
        """Initialize the collector."""
        self.parent = parent
        self.accept = accept

    def collect(self) -> Generator[Metric]:
        """Yield each accepted metric."""
        for metric in self.parent.collect():
            # Test coverage sometimes thinks the false case is uncovered,
            # but in practice OpenMetricsViewTests exercises all the cases
            # we care about.
            if self.accept(metric):  # pragma: no branch
                yield metric


def extract_media_type(content_type: str) -> str:
    """Remove the charset from a fully specified Content-Type."""
    return content_type.split(";", 1)[0].strip()


class RendererFormats(StrEnum):
    """Supported output formats."""

    PROMETHEUS = "prometheus"
    OPENMETRICS = "openmetrics"


class BaseTextRenderer(renderers.BaseRenderer):
    """Base class for returning text responses."""

    @override
    def render(
        self,
        data: str,
        accepted_media_type: str | None = None,
        renderer_context: Mapping[str, Any] | None = None,
    ) -> str:
        """Render data to bytes."""
        accepted_media_type, renderer_context  # fake usage for vulture
        assert self.charset
        return smart_str(data, encoding=self.charset)


class PrometheusRenderer(BaseTextRenderer):
    """Renderer for Prometheus native metrics."""

    media_type = extract_media_type(prometheus.CONTENT_TYPE_LATEST)
    format = RendererFormats.PROMETHEUS


class OpenMetricsRenderer(BaseTextRenderer):
    """Renderer for OpenMetrics format metrics."""

    media_type = extract_media_type(openmetrics.CONTENT_TYPE_LATEST)
    format = RendererFormats.OPENMETRICS


class OpenMetricsView(BaseAPIView):
    """View used to get the OpenMetrics statistics."""

    renderer_classes = [PrometheusRenderer, OpenMetricsRenderer]

    def get(self, request: Request) -> Response:
        """Return OpenMetrics statistics."""
        self.enforce(anonymous_access)
        # Registry of metrics collected separately by each process.
        non_database_registry = FilteredCollector(
            RedisCollector(None),
            lambda metric: metric.name.startswith(("django_", "debusine_")),
        )

        # Registry of metrics based on the database.
        database_registry = CollectorRegistry(auto_describe=True)
        database_registry.register(DebusineCollector())

        match request.accepted_renderer.format:
            case RendererFormats.PROMETHEUS:
                data = prometheus.generate_latest(non_database_registry)
                assert not data or data.endswith(b"\n")
                data += prometheus.generate_latest(database_registry)
                content_type = prometheus.CONTENT_TYPE_LATEST
            case RendererFormats.OPENMETRICS:
                data = openmetrics.generate_latest(non_database_registry)  # type: ignore[no-untyped-call]
                data = data.removesuffix(b"# EOF\n")
                assert not data or data.endswith(b"\n")
                data += openmetrics.generate_latest(database_registry)  # type: ignore[no-untyped-call]
                content_type = openmetrics.CONTENT_TYPE_LATEST
            case _:  # pragma: no cover
                raise AssertionError("An unexpected renderer was requested")
        return Response(
            data, status=status.HTTP_200_OK, content_type=content_type
        )


class OpenMetricsEmitView(BaseAPIView):
    """Emit an additional metric from a separate worker."""

    permission_classes = [IsWorkerTokenAuthenticated]

    def post(self, request: Request) -> Response:
        """Emit a metric from a worker."""
        serializer = EmitMetricSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # Already validated by IsWorkerTokenAuthenticated
        assert isinstance(self.request.auth, Token)
        assert self.request.auth.worker
        if self.request.auth.worker.worker_type not in (
            WorkerType.CELERY,
            WorkerType.SCHEDULER,
        ):
            raise DebusineAPIException(
                title="Error",
                detail=(
                    f"Endpoint is not accessible to "
                    f"{self.request.auth.worker.worker_type} workers"
                ),
                status_code=status.HTTP_403_FORBIDDEN,
            )
        self.enforce(anonymous_access)  # restricted to workers

        metric_type = serializer.validated_data["metric_type"]
        name = serializer.validated_data["name"]
        labels = serializer.validated_data["labels"]
        value = serializer.validated_data["value"]

        # Make sure emit_metric doesn't recursively call back to this view.
        assert not current_task

        try:
            emit_metric(
                metric_type=metric_type, name=name, labels=labels, value=value
            )
        except MetricNotFound as e:
            raise DebusineAPIException(
                title="Metric not found",
                detail=str(e),
                status_code=status.HTTP_404_NOT_FOUND,
            )
        except BadMetricType as e:
            raise DebusineAPIException(
                title="Bad metric type",
                detail=str(e),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        return Response(status=status.HTTP_204_NO_CONTENT)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test for the main entry point of debusine."""

from unittest import TestCase, mock

from debusine.client.__main__ import entry_point


class ConfigHandlerTests(TestCase):
    """Tests for ConfigHandler class."""

    def test_main_entry_point(self) -> None:
        """Ensure client's main entry point calls Cli().parse()."""
        with mock.patch(
            "debusine.client.cli.Cli.execute", autospec=True
        ) as mocked_main:
            entry_point()

        self.assertEqual(mocked_main.call_count, 1)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for scopes."""

from functools import cached_property
from typing import Any, override

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils.html import format_html

from debusine.db.context import context
from debusine.db.models import Scope
from debusine.web.icons import Icons
from debusine.web.views.places import (
    ContainerPlace,
    ResourcePlace,
    ResourcePlaceKwargs,
)
from debusine.web.views.ui.base import UI


class ScopePlace(ResourcePlace):
    """
    Place with a custom URL.

    This subclass tweaks rendering to use images from static files instead of
    bootstrap icons.
    """

    @override
    def as_icon(self, **kwargs: Any) -> str:
        from django.contrib.staticfiles.storage import staticfiles_storage

        url = staticfiles_storage.url(self.icon)
        return format_html(
            """<img src="{url}" style="height: 1em;" """
            """ class="d-inline-block align-middle"> """,
            url=url,
        )


class ScopeUI(UI[Scope]):
    """UI helpers for Scope instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this scope."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.instance.name):
            url = reverse(
                "scopes:detail",
            )

        kwargs: ResourcePlaceKwargs = {
            "title": self.instance.name,
            "breadcrumb": self.instance.name,
            "url": url,
            "icon": Icons.SCOPE,
            "description": f"{self.instance.name} scope",
        }

        # Scope has an icon attribute which may be set to the appropriate
        # logo. If set, use a custom place implementation so we generate icons
        # as <img> tags instead of <span> ones
        if self.instance.icon:
            kwargs["icon"] = self.instance.icon
            return ScopePlace(**kwargs)
        else:
            return ResourcePlace(**kwargs)

    @cached_property
    def place_group_list(self) -> ContainerPlace:
        """Return a place for the current user group in this scope."""
        from debusine.server.scopes import urlconf_scope

        if not (user := self.request.user).is_authenticated:
            raise ImproperlyConfigured(
                "ScopeUI.place_group_list invoked by an anonymous user"
            )

        with urlconf_scope(self.instance.name):
            url = reverse(
                "user:groups",
                kwargs={"username": user.username},
            )

        return ContainerPlace(
            title=f"Group list for {user} in scope {context.scope}",
            breadcrumb="groups",
            url=url,
            description=f"Groups in scope {self.instance.name}"
            f" accessible by {user}",
            parent=self.instance.ui(self.request).place,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Routes for the server application."""

from typing import TYPE_CHECKING, cast

from django.urls import path

from debusine.server import consumers

if TYPE_CHECKING:
    from channels.routing import _ExtendedURLPattern

    _ExtendedURLPattern  # fake usage for vulture

websocket_urlpatterns = [
    cast(
        "_ExtendedURLPattern",
        path(
            "api/ws/1.0/worker/connect/",
            # TODO: Is this a bug in django-stubs?  as_asgi returns a callable
            # returning None, not HTTPResponseBase.
            consumers.WorkerConsumer.as_asgi(),  # type: ignore[arg-type]
            name='ws-connect',
        ),
    ),
    cast(
        "_ExtendedURLPattern",
        path(
            "api/ws/1.0/work-request/on-completed/",
            # TODO: Is this a bug in django-stubs?  as_asgi returns a callable
            # returning None, not HTTPResponseBase.
            consumers.OnWorkRequestCompletedConsumer.as_asgi(),  # type: ignore[arg-type]  # noqa: E501
            name="ws-work-request-on-completed",
        ),
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the GenerateKey task."""

import base64
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from cryptography.x509 import load_pem_x509_certificate
from django.test import TransactionTestCase, override_settings
from nacl.public import PrivateKey, SealedBox

import debusine.worker.tags as wtags
from debusine.artifacts import WorkRequestDebugLogs
from debusine.artifacts.models import EmptyArtifactData, WorkRequestResults
from debusine.assets import AssetCategory, KeyPurpose, SigningKeyData
from debusine.signing.db.models import Key
from debusine.signing.gnupg import gpg_ephemeral_context
from debusine.signing.tasks import GenerateKey
from debusine.signing.tasks.tests.helpers import SigningTaskTestCase
from debusine.test import TestCase


class GenerateKeyTests(
    SigningTaskTestCase[GenerateKey], TestCase, TransactionTestCase
):
    """Test the :py:class:`GenerateKey` task."""

    SAMPLE_TASK_DATA = {"purpose": "uefi", "description": "A UEFI key"}

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "signing:generatekey:version:1"],
        )

    def test_uefi(self) -> None:
        """Integration test: generate and store a real UEFI key."""
        service_private_key = PrivateKey.generate()
        debusine_mock = self.mock_debusine(self.task)

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        # The private key was encrypted using the configured service-wide
        # private key, and the public key has the correct subject.
        key = Key.objects.get(purpose=Key.Purpose.UEFI)
        self.assertEqual(
            load_pem_x509_certificate(
                bytes(key.public_key)
            ).subject.rfc4514_string(),
            "CN=A UEFI key",
        )
        self.assertEqual(key.private_key["storage"], "nacl")
        self.assertEqual(
            base64.b64decode(key.private_key["public_key"].encode()),
            bytes(service_private_key.public_key),
        )
        SealedBox(service_private_key).decrypt(
            base64.b64decode(key.private_key["encrypted"].encode())
        )

        # The task created an asset with the expected properties.
        debusine_mock.asset_create.assert_called_once_with(
            category=AssetCategory.SIGNING_KEY,
            data=SigningKeyData(
                purpose=KeyPurpose.UEFI,
                fingerprint=key.fingerprint,
                public_key=base64.b64encode(key.public_key).decode(),
                description="A UEFI key",
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )

    def test_openpgp_upload(self) -> None:
        """Integration test: generate and store a real OpenPGP upload key."""
        self.task_data = {
            "purpose": "openpgp",
            "description": "An OpenPGP upload key",
        }
        service_private_key = PrivateKey.generate()
        debusine_mock = self.mock_debusine(self.task)

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        # The private key was encrypted using the configured service-wide
        # private key, and the public key has the correct UID.
        key = Key.objects.get(purpose=Key.Purpose.OPENPGP_UPLOAD)
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            result = ctx.key_import(bytes(key.public_key))
            assert getattr(result, "imported", 0) == 1
            gpg_key = ctx.get_key(result.imports[0].fpr)
            self.assertEqual(gpg_key.uids[0].name, "An OpenPGP upload key")
        self.assertEqual(key.private_key["storage"], "nacl")
        self.assertEqual(
            base64.b64decode(key.private_key["public_key"].encode()),
            bytes(service_private_key.public_key),
        )
        SealedBox(service_private_key).decrypt(
            base64.b64decode(key.private_key["encrypted"].encode())
        )

        # The task created a signing asset with the expected properties.
        debusine_mock.asset_create.assert_called_once_with(
            category=AssetCategory.SIGNING_KEY,
            data=SigningKeyData(
                purpose=KeyPurpose.OPENPGP_UPLOAD,
                fingerprint=key.fingerprint,
                public_key=base64.b64encode(key.public_key).decode(),
                description="An OpenPGP upload key",
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )

    def test_openpgp_repository(self) -> None:
        """Integration test: generate and store a real OpenPGP repo key."""
        self.task_data = {
            "purpose": "openpgp-repository",
            "description": "An OpenPGP repository key",
        }
        service_private_key = PrivateKey.generate()
        debusine_mock = self.mock_debusine(self.task)

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        # The private key was encrypted using the configured service-wide
        # private key, and the public key has the correct UID.
        key = Key.objects.get(purpose=Key.Purpose.OPENPGP_REPOSITORY)
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            result = ctx.key_import(bytes(key.public_key))
            assert getattr(result, "imported", 0) == 1
            gpg_key = ctx.get_key(result.imports[0].fpr)
            self.assertEqual(gpg_key.uids[0].name, "An OpenPGP repository key")
        self.assertEqual(key.private_key["storage"], "nacl")
        self.assertEqual(
            base64.b64decode(key.private_key["public_key"].encode()),
            bytes(service_private_key.public_key),
        )
        SealedBox(service_private_key).decrypt(
            base64.b64decode(key.private_key["encrypted"].encode())
        )

        # The task created a signing asset with the expected properties.
        debusine_mock.asset_create.assert_called_once_with(
            category=AssetCategory.SIGNING_KEY,
            data=SigningKeyData(
                purpose=KeyPurpose.OPENPGP_REPOSITORY,
                fingerprint=key.fingerprint,
                public_key=base64.b64encode(key.public_key).decode(),
                description="An OpenPGP repository key",
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )

    def test_generate_failure(self) -> None:
        """If key generation fails, the task does not upload a signing key."""
        debusine_mock = self.mock_debusine(self.task)

        with mock.patch.object(
            Key.objects,
            "generate",
            side_effect=Exception("Key generation failed"),
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.FAILURE)

        # No key was generated.
        self.assertQuerySetEqual(Key.objects.all(), [])

        # No asset was created
        debusine_mock.asset_create.assert_not_called()

        # Debug logs artifact was uploaded
        debusine_mock.upload_artifact.assert_called_once_with(
            WorkRequestDebugLogs(
                category=WorkRequestDebugLogs._category,
                data=EmptyArtifactData(),
                files={
                    "stages.log": self.task.log_directory / "stages.log",
                    "cmd-output.log": (
                        self.task.log_directory / "cmd-output.log"
                    ),
                    "execution.log": (
                        self.task.log_directory / "execution.log"
                    ),
                },
                content_types={
                    "stages.log": "text/plain; charset=utf-8",
                    "cmd-output.log": "text/plain; charset=utf-8",
                    "execution.log": "text/plain; charset=utf-8",
                },
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebianPipeline work request plugin."""

from typing import ClassVar, cast, override

from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.debian_pipeline import (
    DebianPipelineViewWorkRequestPlugin,
)
from debusine.web.views.qa import QATab
from debusine.web.views.work_request import Tab


class DebianPipelineViewWorkRequestPluginTests(TestCase):
    """Tests for :py:class:`DebianPipelineViewWorkRequestPlugin`."""

    scenario = scenarios.UIRegressionTesting(simulate=False)

    pipeline_without_qa: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.pipeline_without_qa = cls.playground.create_workflow(
            cls.scenario.template_upload_to_sid,
            task_data={
                "source_artifact": cls.scenario.new_source.pk,
                "architectures": ["all", "amd64", "i386"],
                "enable_regression_tracking": True,
            },
            workspace=cls.scenario.workspace,
        )

    def plugin(
        self, work_request: WorkRequest | None = None
    ) -> DebianPipelineViewWorkRequestPlugin:
        if work_request is None:
            work_request = self.scenario.workflow_debian_pipeline
        return DebianPipelineViewWorkRequestPlugin(work_request.get_task())

    def test_enabled(self) -> None:
        plugin = self.plugin()
        self.assertTrue(plugin.is_enabled())

    def test_disabled_no_qa(self) -> None:
        plugin = self.plugin(self.pipeline_without_qa)
        self.assertFalse(plugin.is_enabled())

    def test_disabled_no_qa_data(self) -> None:
        self.scenario.workflow_qa.output_data = None
        self.scenario.workflow_qa.save()
        plugin = self.plugin()
        self.assertFalse(plugin.is_enabled())

    def test_add_tabs(self) -> None:
        plugin = self.plugin()

        tabs: list[Tab | int] = [1, 2, 3]
        plugin.add_tabs(cast(list[Tab], tabs))
        self.assertEqual(len(tabs), 4)
        self.assertEqual(tabs[0], 1)
        self.assertEqual(tabs[2:], [2, 3])
        tab = tabs[1]
        assert isinstance(tab, QATab)
        self.assertEqual(tab.work_request, self.scenario.workflow_qa)

    def test_add_tabs_disabled(self) -> None:
        plugin = self.plugin(self.pipeline_without_qa)

        tabs: list[int] = [1, 2, 3]
        plugin.add_tabs(cast(list[Tab], tabs))
        self.assertEqual(tabs, [1, 2, 3])

    def test_get_context_data(self) -> None:
        plugin = self.plugin()
        self.assertEqual(plugin.get_context_data(), {})
